import sys
from typing import Union, Counter, Tuple

import numpy as np
from sklearn.preprocessing import LabelBinarizer


def simple_majority_voting(array: np.ndarray):
    if array.size == 0:
        return None
    c = Counter(array)
    return c.most_common(1)[0][0]


def shuffle_in_unison(*arrays: np.ndarray, seed: int | None = None, axis: int = 0):
    """
    Shuffle any number of arrays with the same random permutation along the given axis.
    Returns shuffled copies (does not modify inputs) and the permutation used.

    All arrays must have the same length along `axis`.
    """
    if not arrays:
        raise ValueError("At least one array must be provided.")

    length = arrays[0].shape[axis]
    for i, arr in enumerate(arrays[1:], start=1):
        if arr.shape[axis] != length:
            raise ValueError(
                f"Length mismatch along axis {axis}: "
                f"array[0] has {length}, but array[{i}] has {arr.shape[axis]}"
            )

    rng = np.random.default_rng(seed)
    perm = rng.permutation(length)
    shuffled = [np.take(arr, perm, axis=axis) for arr in arrays]
    return *shuffled, perm


def _unison_shuffle_np_arrays_3(a: np.ndarray, b: np.ndarray, c: np.ndarray, seed: int | None = None):
    """Shuffle three arrays in unison along axis 0 and return the shuffled arrays.
    This is a thin wrapper around shuffle_in_unison to match existing expectations.
    """
    a_s, b_s, c_s, _ = shuffle_in_unison(a, b, c, seed=seed, axis=0)
    return a_s, b_s, c_s


class Dataset:
    # Data to train on. In many cases this is considered to be equivalent to the enrolment data, but for distance
    # learning, these sets can be different.
    _trainset_data: np.ndarray = None
    _trainset_labels: np.ndarray = None

    # Data to validate on. Will be set to the same as the training data if not specified otherwise.
    _validationset_data: np.ndarray = None
    _validationset_labels: np.ndarray = None

    # Actual biometric enrollment data. Will be set to the training dawta if not specified otherwise.
    _testset_enrollment_data: np.ndarray = None
    _testset_enrollment_labels: np.ndarray = None

    # ACtual biometric matching (testset) data.
    _testset_matching_data: np.ndarray = None
    _testset_matching_labels: np.ndarray = None

    def __init__(self,
                 trainset_data: np.ndarray,
                 trainset_labels: np.ndarray,
                 validationset_data: np.ndarray = None,
                 validationset_labels: np.ndarray = None,
                 testset_enrollment_data: np.ndarray = None,
                 testset_enrollment_labels: np.ndarray = None,
                 testset_matching_data: np.ndarray = None,
                 testset_matching_labels: np.ndarray = None,
                 normalization_check: bool = False):
        self._trainset_data = self._check_and_convert_type_data(trainset_data, name="trainset_data")
        self._trainset_labels = self._check_and_convert_type_label(trainset_labels, name="trainset_labels")
        self._validationset_data = self._check_and_convert_type_data(validationset_data, name="validationset_data")
        self._validationset_labels = self._check_and_convert_type_label(validationset_labels,
                                                                       name="validationset_labels")
        self._testset_enrollment_data = self._check_and_convert_type_data(testset_enrollment_data,
                                                                         name="testset_enrollment_data")
        self._testset_enrollment_labels = self._check_and_convert_type_label(testset_enrollment_labels,
                                                                            name="testset_enrollment_labels")
        self._testset_matching_data = self._check_and_convert_type_data(testset_matching_data,
                                                                       name="testset_matching_data")
        self._testset_matching_labels = self._check_and_convert_type_label(testset_matching_labels,
                                                                          name="testset_matching_labels")

        # perform checks of data
        self._check_data_pairs_provided()  # check if corresponding pairs were provided,
        # no array is empty and corresponding pairs are of the same length
        self._check_nan()  # check if any NaNs exists
        self._check_dataset_is_balanced()
        if normalization_check:
            self._normalisation_check()  # check if data is normalized in [-1, 1]

        self._setup_missing_sets()
        self._setup_reject_label()

        # OHC Encoding
        self.ohc_encoder = LabelBinarizer()
        self.ohc_encoder.fit(self._gather_labels())

    def _setup_missing_sets(self):
        if self._validationset_data is None:
            if self._testset_enrollment_data is None:
                raise ValueError(f"Critical: a validation set needs to be present.")
            else:
                self._validationset_data = self._testset_enrollment_data
                self._validationset_labels = self._testset_enrollment_labels
                print(f"Warning: because the validation set is None, the supplied testset enrollment data will "
                      f"be used instead.", file=sys.stderr)
        if self._testset_enrollment_data is None:
            if self._testset_matching_data is None:
                print(f"Warning: testset enrollment and matching data are None. This will cause all evaluation "
                      f"methods to fail.", file=sys.stderr)
            else:
                self._testset_enrollment_data = self._testset_matching_data
                self._testset_enrollment_labels = self._testset_matching_labels
                print(f"Warning: because the enrollment testset is None, the matching testset will be used "
                      f"instead.", file=sys.stderr)
        else:
            if self._testset_matching_data is None:
                print(f"Warning: testset matching data is None. This will cause all evaluation "
                      f"methods to fail.", file=sys.stderr)

    def _setup_reject_label(self):
        self.reject_label = 'reject'  # In some cases this will be changed further down in the method
        if self._testset_matching_labels is not None:
            try:
                match_label_set = set(self._testset_matching_labels)
                enroll_label_set = set(self._testset_enrollment_labels)
            except TypeError:  # np.ndarray is not hashable
                match_label_set = set(self._testset_matching_labels.tolist())
                enroll_label_set = set(self._testset_enrollment_labels.tolist())
            diff = match_label_set.difference(enroll_label_set)
            if len(diff) == 1:
                self.reject_label = diff.pop()
            else:
                for user_to_reject in diff:  # This does nothing if len(diff) == 0, but it also doesn't hurt
                    self._testset_matching_labels[self._testset_matching_labels == user_to_reject] = self.reject_label

    def view_one_hot_encoded_labels(self):
        def convert_to_onehot(labels: np.ndarray):
            return None if labels is None else self._label_binarizer.transform(labels)

        return convert_to_onehot(self._trainset_labels), convert_to_onehot(self._validationset_labels), \
               convert_to_onehot(self._testset_enrollment_labels), convert_to_onehot(self._testset_matching_labels)

    @staticmethod
    def _check_and_convert_type_data(obj, name) -> np.array:
        """Check the types of objects, print warnings and perform conversions."""
        return Dataset._check_and_convert_type(obj, name, 'float32')

    @staticmethod
    def _check_and_convert_type_label(obj, name) -> np.array:
        """Check the types of objects, print warnings and perform conversions."""
        result = Dataset._check_and_convert_type(obj, name, str)
        if result is not None and len(result.shape) != 1:
            raise ValueError("Critical: label arrays must be one-dimensional.")
        return result

    @staticmethod
    def _check_and_convert_type(obj, name, dtype):
        if obj is None:
            return None
        if not isinstance(obj, np.ndarray):
            print(f"Warning: provided object {name} is of type {str(type(obj))} (expected: np.ndarray). "
                  f"Attempting a cast to np.ndarray.", file=sys.stderr)
            result = np.array(obj, dtype=dtype)  # this fails if dtype is float and the resulting array would be ragged
        elif obj.dtype != dtype:
            print(f"Warning: provided object {name} contains values of type {obj.dtype} (expected: {dtype}). "
                  f"Attempting a cast to expected datatype!", file=sys.stderr)
            result = obj.astype(dtype=dtype, copy=False)
        else:
            result = obj
        return result

    def _check_data_pairs_provided(self):
        """Checks that if training data is provided also labels are provided. Raises exception if not."""

        # check for None
        if self._trainset_data is None or len(self._trainset_data) == 0:
            raise ValueError("Critical: _trainset_data must not be None or empty.")

        if self._trainset_labels is None:
            raise ValueError("Critical: _trainset_labels must not be None.")

        def _check(data, labels, name):
            if data is not None:
                if labels is None:
                    raise ValueError(f"Critical: if {name}_data is not None, {name}_labels must also be not None"
                                     f"However, {name}_labels is None! You rascal!")
                if len(data) == 0:
                    raise ValueError(f"Critical: length of {name}_data is {len(self._trainset_labels)}, "
                                     f"but must be greater than 0!")
                if len(data) != len(labels):
                    raise ValueError(f"Critical: length of {name}_data ({len(self._trainset_data)}) is not equal to the "
                                     f"length of {name}_labels ({len(self._trainset_labels)})!")

        _check(self._validationset_data, self._validationset_labels, 'validationset')
        _check(self._testset_enrollment_data, self._testset_enrollment_labels, 'testset_enrollment')
        _check(self._testset_matching_data, self._testset_matching_labels, 'testset_matching')

    def _check_nan(self):
        def _check_nan(array: np.ndarray, name):
            try:
                if array is not None and np.isnan(array).any():
                    raise ValueError(f'Critical: provided array {name} contains NaN values!')
            except TypeError:
                print(f'Warning: could not check if NaNs exist in array {name}, as its dtype is "{array.dtype}". '
                      f'A numerical dtype (e.g., int or float) is expected. This might lead to subsequent '
                      f'critical errors and wrong calculations.', file=sys.stderr)

        _check_nan(self._trainset_data, name="_trainset_data")
        _check_nan(self._testset_matching_data, name="_testset_matching_data")
        _check_nan(self._testset_enrollment_data, name="_testset_enrollment_data")
        _check_nan(self._validationset_data, name="_validationset_data")

    def _check_dataset_is_balanced(self):
        return True  # TODO implement

    def _normalisation_check(self):
        def _normalisation_check(array: np.ndarray):
            if array is not None:
                _max, _min = np.amax(array), np.amin(array)
                if _max > 1:
                    raise ValueError(f"Critical: dataset contains maximum value {_max} which exceeds [-1, 1].")
                if _min < -1:
                    raise ValueError(f"Critical: dataset contains minimum value {_min} which exceeds [-1, 1].")

        _normalisation_check(self._trainset_data)
        _normalisation_check(self._validationset_data)
        _normalisation_check(self._testset_enrollment_data)
        _normalisation_check(self._testset_matching_data)

    def _gather_labels(self) -> np.ndarray:
        """Concatenates all labels."""
        labels = [self._trainset_labels]
        if self._validationset_labels is not None:
            labels.append(self._validationset_labels)
        if self._testset_enrollment_labels is not None:
            labels.append(self._testset_enrollment_labels)
        if self._testset_matching_labels is not None:
            labels.append(self._testset_matching_labels)
        return np.concatenate(labels)

    def _get_unique_labels(self):
        return np.unique(self._gather_labels())

    def get_unique_classes(self):
        return np.unique(self._gather_labels())

    def get_unique_enrollment_classes(self):
        return np.unique(self._testset_enrollment_labels)

    def get_unique_matching_classes(self):
        return np.unique(self._testset_matching_labels)

    def _get_enrollment_data_filter(self, filter):
        """
        Return enrollment data filtered by the given filter interacting with enrollment labels.

        The filter can be one of the following:
          - callable: predicate taking a label and returning True/False
          - scalar label value: selects items equal to this label
          - iterable of label values: selects items whose label is in the iterable
        """
        if self._testset_enrollment_data is None or self._testset_enrollment_labels is None:
            raise ValueError("Enrollment data or labels are not available to filter.")

        labels = self._testset_enrollment_labels

        # Build boolean mask according to provided filter
        mask: np.ndarray
        if callable(filter):
            # Apply predicate per label
            mask = np.array([bool(filter(lbl)) for lbl in labels])
        else:
            try:
                # If filter is iterable (e.g., list, set), treat as membership filter
                values = set(list(filter))  # may raise TypeError if not iterable
                mask = np.array([lbl in values for lbl in labels])
            except TypeError:
                # Treat as scalar equality
                mask = (labels == filter)

        return self._testset_enrollment_data[mask]

    def _get_matching_data_filter(self, filter):
        """
        Return matching data filtered by the given filter interacting with matching labels.

        The filter can be one of the following:
          - callable: predicate taking a label and returning True/False
          - scalar label value: selects items equal to this label
          - iterable of label values: selects items whose label is in the iterable
        """
        if self._testset_matching_data is None or self._testset_matching_labels is None:
            raise ValueError("Matching data or labels are not available to filter.")

        labels = self._testset_matching_labels

        # Build boolean mask according to provided filter
        if callable(filter):
            mask = np.array([bool(filter(lbl)) for lbl in labels])
        else:
            try:
                values = set(list(filter))
                mask = np.array([lbl in values for lbl in labels])
            except TypeError:
                mask = (labels == filter)

        return self._testset_matching_data[mask]

    def get_dl_training_shape(self):
        return self._testset_enrollment_data.shape[1:]

    def get_enrollment_labels(self, ohc=False):
        if ohc:
            return self.ohc_encoder.transform(self._testset_enrollment_labels)
        else:
            return self._testset_enrollment_labels

    def get_matching_labels(self, ohc=False):
        if ohc:
            return self.ohc_encoder.transform(self._testset_matching_labels)
        else:
            return self._testset_matching_labels

    def get_enrollment_data(self, filter=None):
        if filter is not None:
            return self._get_enrollment_data_filter(filter)
        else:
            return self._testset_enrollment_data

    def get_trainset_data(self):
        return self._trainset_data

    def get_trainset_labels(self):
        return self._trainset_labels

    def get_matching_data(self, filter=None):
        if filter is not None:
            return self._get_matching_data_filter(filter)
        else:
            return self._testset_matching_data

    def ohc_labels_to_classes(self, ohc_labels):
        return self.ohc_encoder.inverse_transform(ohc_labels)


class WindowslicedDataset(Dataset):
    def __init__(self,
                 window_slicing_stride: int,
                 window_slicing_size: int,
                 trainset_data: Union[list, np.ndarray],
                 trainset_labels: Union[list, np.ndarray],
                 validationset_data: Union[list, np.ndarray] = None,
                 validationset_labels: Union[list, np.ndarray] = None,
                 testset_enrollment_data: Union[list, np.ndarray] = None,
                 testset_enrollment_labels: Union[list, np.ndarray] = None,
                 testset_matching_data: Union[list, np.ndarray] = None,
                 testset_matching_labels: Union[list, np.ndarray] = None,
                 normalization_check: bool = False,
                 shuffle_window_slices: bool = False,
                 voting_function = None,
                 ):
        if voting_function is None:
            self.voting_function = simple_majority_voting

        self._trainset_data: np.ndarray = trainset_data  # type hints here are just for IDE support
        self._trainset_labels: np.ndarray = trainset_labels
        self._validationset_data: np.ndarray = validationset_data
        self._validationset_labels: np.ndarray = validationset_labels
        self._testset_enrollment_data: np.ndarray = testset_enrollment_data
        self._testset_enrollment_labels: np.ndarray = testset_enrollment_labels
        self._testset_matching_data: np.ndarray = testset_matching_data
        self._testset_matching_labels: np.ndarray = testset_matching_labels
        self._setup_missing_sets()  # applying the replacement rules for missing sets in this way seems a little


        def _len_or_none(data: np.ndarray):
            return None if data is None else len(data)

        trainset_origin_len = _len_or_none(self._trainset_data)
        validationset_origin_len = _len_or_none(self._validationset_data)
        testset_enrollment_origin_len = _len_or_none(self._testset_enrollment_data)
        testset_matching_origin_len = _len_or_none(self._testset_matching_data)

        _trainset_data, _trainset_labels, self.trainset_slicedsample_origin = self._transform_for_windowslicing(
            self._trainset_data, self._trainset_labels, window_slicing_stride, window_slicing_size, shuffle_window_slices)

        _validationset_data, _validationset_labels, self.validationset_slicedsample_origin = \
            self._transform_for_windowslicing(self._validationset_data, self._validationset_labels, window_slicing_stride,
                                              window_slicing_size, shuffle_window_slices)

        _testset_enrollment_data, _testset_enrollment_labels, self.testset_enrollment_slicedsample_origin = \
            self._transform_for_windowslicing(self._testset_enrollment_data, self._testset_enrollment_labels,
                                              window_slicing_stride, window_slicing_size, shuffle_window_slices)

        _testset_matching_data, _testset_matching_labels, self.testset_matching_slicedsample_origin = \
            self._transform_for_windowslicing(self._testset_matching_data, self._testset_matching_labels,
                                              window_slicing_stride, window_slicing_size, shuffle_window_slices)

        self._create_origin_masks(trainset_origin_len, validationset_origin_len, testset_enrollment_origin_len,
                                  testset_matching_origin_len)

        super().__init__(_trainset_data, _trainset_labels, _validationset_data, _validationset_labels,
                         _testset_enrollment_data, _testset_enrollment_labels, _testset_matching_data,
                         _testset_matching_labels, normalization_check)

    def _create_origin_masks(self, trainset_origin_len, validationset_origin_len, testset_enrollment_origin_len,
                             testset_matching_origin_len):
        def _create_origin_mask(origin_len, slicedsample_origin):
            if slicedsample_origin is None:
                return None
            else:
                result_origin_mask = []
                for origin_id in range(origin_len):
                    result_origin_mask.append(slicedsample_origin == origin_id)
                return np.array(result_origin_mask, dtype=bool)

        self.trainset_sliceorigin_mask = _create_origin_mask(trainset_origin_len, self.trainset_slicedsample_origin)
        self.validationset_sliceorigin_mask = _create_origin_mask(validationset_origin_len,
                                                                  self.validationset_slicedsample_origin)
        self.testset_enrollment_sliceorigin_mask = _create_origin_mask(testset_enrollment_origin_len,
                                                                       self.testset_enrollment_slicedsample_origin)
        self.testset_matching_sliceorigin_mask = _create_origin_mask(testset_matching_origin_len,
                                                                     self.testset_matching_slicedsample_origin)

    @staticmethod
    def _transform_for_windowslicing(data: Union[list, np.ndarray],
                                     labels: Union[list, np.ndarray],
                                     window_slicing_stride: int,
                                     window_slicing_size: int,
                                     shuffle_window_slices: bool):
        if data is None:
            return None, None, None

        if not len(data) == len(labels):
            raise ValueError("Critical: data and labels differ in size.")

        ret_sliced_data, ret_sliced_labels, ret_sample_origin_ids = [], [], []

        for origin_id, (sample, label) in enumerate(zip(data, labels)):
            array_slices, label_slices, sample_origin_ids = [], [], []

            lost_samples_cnt, added_cnt = 0, 0

            for step_idx in range(0, len(sample) - window_slicing_size + 1, window_slicing_stride):
                arr = np.array(sample[step_idx:step_idx + window_slicing_size],
                               dtype=float)  # this fails if the resulting array would be ragged

                if not arr.any():
                    pass  # do not append if slice consists only of zero
                else:
                    array_slices.append(arr)
                    label_slices.append(label)
                    sample_origin_ids.append(origin_id)

            for i in range(len(array_slices)):
                if len(array_slices[i]) == window_slicing_size:
                    ret_sliced_data.append(array_slices[i])
                    ret_sliced_labels.append(label_slices[i])
                    ret_sample_origin_ids.append(sample_origin_ids[i])
                    added_cnt += 1
                else:
                    lost_samples_cnt += 1

            if lost_samples_cnt > 0:
                print(f'Warning: during slicing  {lost_samples_cnt} malformed slices needed to be discarded. '
                      f'Check slicing parameters.', file=sys.stderr)

        try:
            ret_sliced_data = np.array(ret_sliced_data, dtype=float)
        except ValueError:
            print("Critical: slicing the data produced slices differing in size.", file=sys.stderr)
            raise

        ret_sliced_labels = np.array(ret_sliced_labels, dtype=str)
        ret_sample_origin_ids = np.array(ret_sample_origin_ids, dtype=int)

        print(f'Info: sliced window data takes up {round(ret_sliced_data.nbytes / 1024 / 1024, 2)}mb.')

        if shuffle_window_slices:
            return _unison_shuffle_np_arrays_3(ret_sliced_data, ret_sliced_labels, ret_sample_origin_ids)
        else:
            return ret_sliced_data, ret_sliced_labels, ret_sample_origin_ids

    def apply_voting_for_testset_matching_slices(self, predictions_slices: np.ndarray,
                                                 ground_truth_slices: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        ground_truth = np.zeros((len(self.testset_matching_sliceorigin_mask),), dtype=ground_truth_slices.dtype)
        predictions = np.zeros((len(self.testset_matching_sliceorigin_mask),), dtype=predictions_slices.dtype)

        for sample_id, sample_mask in enumerate(self.testset_matching_sliceorigin_mask):
            if np.any(sample_mask):
                if predictions_slices.ndim == 2:  # ohc tabular predictions
                    predictions[sample_id] = self.voting_function(np.argmax(predictions_slices[sample_mask], axis=1))
                    ground_truth[sample_id] = np.argmax(ground_truth_slices[sample_mask], axis=1)[0]
                else:
                    predictions[sample_id] = self.voting_function(predictions_slices[sample_mask])
                    ground_truth[sample_id] = ground_truth_slices[sample_mask][0]  # assuming consistency within slices
            else:
                predictions[sample_id] = None  # Default value for predictions when no slices match
                ground_truth[sample_id] = None  # Default value for ground truth when no slices match

        return predictions, ground_truth



import numpy as np

from useckit.paradigms.distance_learning.prediction_models.distance_prediction_model_base import \
    DistanceBasePredictionModel


def _create_all_distance_pairs(p_array_1: np.ndarray, p_array_2: np.ndarray,
                               distance_metric: DistanceBasePredictionModel,
                               tradeoff_computation_speed_for_memory: bool):
    assert p_array_1[0].shape == p_array_2[0].shape  # beyond their length, the arrays need to have the same shape
    if tradeoff_computation_speed_for_memory:
        return _create_distance_pairs_iteration(p_array_1, p_array_2,
                                                distance_metric)
    else:
        return _create_distance_pairs_broadcasting(p_array_1, p_array_2,
                                                   distance_metric)


# This method creates all possible pairs between p1 and p2 and passes them as one to the distance metric. This avoids
# the inefficient streaming that occurs when passing the pairs individually to the distance metric, but it consumes a
# lot more memory.
def _create_distance_pairs_broadcasting(p_array_1: np.ndarray, p_array_2: np.ndarray,
                                        distance_metric: DistanceBasePredictionModel):
    # gather all pairs that have to be predicted before later splitting them back up into the predictions for the
    # individual p_array_1
    p1_broadcast_array_shape = [len(p_array_2) * len(p_array_1)]  # because we compare
    # each sample in p1 with every sample in p2, we need enough space to store as many p1 and p2 sample pairs
    for s in p_array_2.shape[1:]:  # each other dimension needs to be equivalent to a single sample
        p1_broadcast_array_shape.append(s)
    # p1_broadcast_array_shape and p2_broadcast_array hold the two sides of the pairs and thus have the same shape
    p1_broadcast_array = np.zeros(shape=tuple(p1_broadcast_array_shape), dtype=p_array_2.dtype)
    p2_broadcast_array = np.zeros(shape=tuple(p1_broadcast_array_shape), dtype=p_array_2.dtype)
    p1_original_indices = np.zeros(
        shape=(len(p_array_2) * len(p_array_1),), dtype=int)  # this will contain the original
    # indices of the broadcasted p_array_1
    broadcast_index = 0
    for index, sample in enumerate(p_array_1):
        sample_broadcast = np.broadcast_to(sample, shape=p_array_2.shape)
        indices = np.ones(shape=(len(p_array_2),)) * index
        from_i = broadcast_index
        to_i = broadcast_index + len(sample_broadcast)

        p1_broadcast_array[from_i:to_i] = sample_broadcast
        p2_broadcast_array[from_i:to_i] = p_array_2
        p1_original_indices[from_i:to_i] = indices

        broadcast_index = to_i
    distances = distance_metric.predict(p1_broadcast_array, p2_broadcast_array)
    return distances, p1_original_indices


# This method gathers batches of len(p_array_2) many pairs and calculates the distances between them. This strikes a
# balance between the computationally fastest method above and memory efficiency,
def _create_distance_pairs_iteration(p_array_1: np.ndarray, p_array_2: np.ndarray,
                                     distance_metric: DistanceBasePredictionModel):
    distances = np.zeros(
        shape=(len(p_array_2) * len(p_array_1),))  # result
    p1_original_indices = np.zeros(
        shape=(len(p_array_2) * len(p_array_1),), dtype=int)  # this will contain the original
    # indices of the broadcasted p_array_1
    broadcast_index = 0
    for index, sample in enumerate(p_array_1):
        sample_broadcast = np.broadcast_to(sample, shape=p_array_2.shape)
        indices = np.ones(shape=(len(p_array_2),)) * index
        from_i = broadcast_index
        to_i = broadcast_index + len(sample_broadcast)

        sample_distances = distance_metric.predict(sample_broadcast, p_array_2)
        distances[from_i:to_i] = sample_distances
        p1_original_indices[from_i:to_i] = indices

        broadcast_index = to_i
    return distances, p1_original_indices



import numpy as np

from useckit.util.dataset import Dataset
from useckit.util.dataset_windowsliced import WindowslicedDataset
from useckit.util.utils import make_some_intelligent_noise as make_some_intelligent_noise_util


def make_some_intelligent_noise(labels: int = 4, shape: tuple = (100, 100, 100), noisiness: float = 0.1):
    biases = np.array([i / labels for i in range(labels)]) + (1 / (2 * labels))
    x = []
    y = []
    shape = list(shape)
    shape[0] = int(shape[0] / labels)
    i = 0
    for b in biases:
        x.extend(np.random.normal(b, noisiness, shape))
        y.extend(np.ones((shape[0],)) * i)
        i += 1
    x = np.array(x)
    y = np.array(y)
    x = np.clip(x, -1, 1)
    return x, y


def make_dataset(*args, **kwargs):
    x_train, y_train, x_val, y_val, x_test_enroll, y_test_enroll, x_test_match, y_test_match = \
        _make_data(*args, **kwargs)
    return Dataset(x_train, y_train, x_val, y_val, x_test_enroll, y_test_enroll, x_test_match, y_test_match)


def make_windowsliced_dataset(window_slicing_stride: int, window_slicing_size: int, *args, **kwargs):
    x_train, y_train, x_val, y_val, x_test_enroll, y_test_enroll, x_test_match, y_test_match = \
        _make_data(*args, **kwargs)
    return WindowslicedDataset(window_slicing_stride, window_slicing_size, x_train, y_train, x_val, y_val,
                               x_test_enroll, y_test_enroll, x_test_match, y_test_match)


def _make_data(*args, **kwargs):
    if 'labels' not in kwargs:
        kwargs['labels'] = 6
    num_reject_labels = int(kwargs['labels'] / 3)
    reject_labels = []
    reject_label_masks = []

    x_train, y_train = make_some_intelligent_noise_util(*args, **kwargs)
    x_val, y_val = make_some_intelligent_noise_util(*args, **kwargs)
    x_test_enroll, y_test_enroll = make_some_intelligent_noise_util(*args, **kwargs)

    # remove the middle most labels to not have edge labels mess with random forests
    start_index = int(len(y_train) - num_reject_labels - (len(y_train) - num_reject_labels) / 2)
    for i in range(start_index, start_index + num_reject_labels, 1):
        if y_train[-i - 1] not in reject_labels:
            reject_labels.append(y_train[-i - 1])
            reject_label_masks.append(y_train == y_train[-i - 1])

    total_reject_label_mask = reject_label_masks[0]
    for mask in reject_label_masks:
        total_reject_label_mask = np.logical_or(total_reject_label_mask, mask)

    total_reject_label_mask = np.invert(total_reject_label_mask)

    x_train = x_train[total_reject_label_mask]
    y_train = y_train[total_reject_label_mask]
    x_val = x_val[total_reject_label_mask]
    y_val = y_val[total_reject_label_mask]
    x_test_enroll = x_test_enroll[total_reject_label_mask]
    y_test_enroll = y_test_enroll[total_reject_label_mask]

    x_test_match, y_test_match = make_some_intelligent_noise_util(*args, **kwargs)
    return x_train, y_train, x_val, y_val, x_test_enroll, y_test_enroll, x_test_match, y_test_match



import numpy as np

from ._fixed_false_negative_thresholding_method import FixedFalseNegativeThresholdingMethod
from .anomaly_evaluation_method_base import AnomalyBaseEvaluationMethod, BaseThresholdingMethod
from ..prediction_models.anomaly_prediction_model_base import AnomalyBasePredictionModel
from ....evaluation.verification import VerificationModel, perform_verification_evaluation
from ....util.dataset import Dataset


class AnomalyVerificationModel(VerificationModel):

    def __init__(self,
                 anomaly_prediction_model: AnomalyBasePredictionModel,
                 rejection_thresholds_per_trained_user: [float]):
        self.anomaly_prediction_model = anomaly_prediction_model
        self.rejection_thresholds_per_trained_user = rejection_thresholds_per_trained_user

    def verify(self, samples: np.ndarray, identity_claims: np.ndarray) -> np.ndarray:
        preds, labels = self.anomaly_prediction_model.predict(samples)

        result = np.zeros(shape=identity_claims.shape)
        for i in range(len(identity_claims)):  # iterate over all samples
            label_index = np.argwhere(labels == identity_claims[i])[0][0]  # this should never give more than 1 value!
            reject = preds[label_index][i] > self.rejection_thresholds_per_trained_user[identity_claims[i]]
            if reject:
                result[i] = 0
            else:
                result[i] = 1
        return result == 1  # dirty transform to bool array


class Verification(AnomalyBaseEvaluationMethod):

    def __init__(self,
                 output_dir: str = "evaluation_verification",
                 threshold_method: BaseThresholdingMethod = None):
        super().__init__(output_dir)
        self.threshold_method = threshold_method

    def evaluate(self, dataset: Dataset, prediction_model: AnomalyBasePredictionModel, **kwargs):
        if self.threshold_method is None:
            self.threshold_method = FixedFalseNegativeThresholdingMethod(prediction_model, self.output_dir)
        thresholds = self.threshold_method.compute_thresholds(dataset.testset_enrollment_data,
                                                              dataset.testset_enrollment_labels)
        verification_model = AnomalyVerificationModel(prediction_model, thresholds)
        perform_verification_evaluation(verification_model, dataset, self.output_dir)



from abc import abstractmethod

import numpy as np

from useckit.evaluation.identification import IdentificationModel, perform_identification_evaluation
from useckit.util.dataset import Dataset


#  This overrides the pure IdentificationModel to avoid code duplication in the actual evaluation method below.
class IdentificationOrRejectModel(IdentificationModel):

    def identify(self, samples: np.ndarray) -> np.ndarray:
        return self.identify_or_reject(samples)

    @abstractmethod
    def identify_or_reject(self, samples: np.ndarray):
        pass


def perform_identification_or_reject_evaluation(identification_model: IdentificationOrRejectModel, dataset: Dataset,
                                                output_folder: str):
    perform_identification_evaluation(identification_model=identification_model, dataset=dataset,
                                      output_folder=output_folder)



from abc import abstractmethod
from useckit.util.dataset import Dataset
from ..._paradigm_base import PredictionModelBase


class DistanceBasePredictionModel(PredictionModelBase):

    def __init__(self, output_dir: str = "model_out", verbose: bool = False):
        super().__init__(output_dir=output_dir, verbose=verbose)

    @abstractmethod
    def fit(self, dataset: Dataset):
        pass

    @abstractmethod
    def predict(self, x_test_1, x_test_2):
        pass



import os
import re
import threading
from datetime import datetime


class UseckitSessionFolder:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls, folder_name: str = None):
        """Implements singleton pattern with thread safety."""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, folder_name: str = None):
        # Ensure __init__ runs only once per singleton creation
        if hasattr(self, "_initialized") and self._initialized:
            return

        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        self.folder_name = folder_name or f"_useckit_{timestamp}"
        self.folder_base_path = os.path.join(os.getcwd(), self.folder_name)

        os.makedirs(self.folder_base_path, exist_ok=True)
        self._initialized = True

    @classmethod
    def get_instance(cls):
        """Access the existing singleton instance."""
        if cls._instance is None:
            raise RuntimeError("UseckitSessionFolder not created yet.")
        return cls._instance

    @classmethod
    def destroy_instance(cls):
        """Destroys the singleton instance (does not delete the folder)."""
        with cls._lock:
            cls._instance = None

    def __repr__(self):
        return f"<UseckitSessionFolder path={self.folder_base_path}>"

    def get_folder_basepath(self):
        return self.folder_base_path


class ExperimentFolder:
    def __init__(self, folder_description: str = None):
        _session = UseckitSessionFolder()
        _base_path = _session.get_folder_basepath()

        # Find all existing experiment folders 
        existing_folders = [
            name for name in os.listdir(_base_path)
            if os.path.isdir(os.path.join(_base_path, name)) and name.startswith("experiment_")
        ]

        # Extract numeric indices from folder names using regex
        numbers = []
        for name in existing_folders:
            match = re.match(r"experiment_(\d+)", name)
            if match:
                numbers.append(int(match.group(1)))

        next_index = (max(numbers) + 1) if numbers else 0

        # Construct new folder name
        if folder_description:
            safe_desc = re.sub(r"[^a-zA-Z0-9_-]", "_", folder_description.strip())
            folder_name = f"experiment_{next_index}_{safe_desc}"
        else:
            folder_name = f"experiment_{next_index}"

        self.folder_path = os.path.join(_base_path, folder_name)

        # Create the folder
        os.makedirs(self.folder_path, exist_ok=True)

    def get_path(self):
        return self.folder_path

    def __repr__(self):
        return f"<ExperimentFolder path={self.folder_path}>"



from abc import abstractmethod, ABC

import numpy as np

from ..prediction_models.anomaly_prediction_model_base import AnomalyBasePredictionModel
from ..._paradigm_base import EvaluationMethodBase
from ....util.dataset import Dataset


class BaseThresholdingMethod(ABC):

    # One threshold for each user the anomaly model was trained on
    @abstractmethod
    def compute_thresholds(self, enrollment_data: np.ndarray, enrollment_labels: np.ndarray) -> [float]:
        pass


class AnomalyBaseEvaluationMethod(EvaluationMethodBase):

    @abstractmethod
    def evaluate(self, dataset: Dataset, prediction_model: AnomalyBasePredictionModel, **kwargs):
        pass



import tensorflow as tf

def _contrastive_loss(y_true, y_pred, margin = 1.0) -> tf.Tensor:
    r"""Computes the contrastive loss between `y_true` and `y_pred`.

    This loss encourages the embedding to be close to each other for
    the samples of the same label and the embedding to be far apart at least
    by the margin constant for the samples of different labels.

    The euclidean distances `y_pred` between two embedding matrices
    `a` and `b` with shape `[batch_size, hidden_size]` can be computed
    as follows:

    >>> a = tf.constant([[1, 2],
    ...                 [3, 4],
    ...                 [5, 6]], dtype=tf.float16)
    >>> b = tf.constant([[5, 9],
    ...                 [3, 6],
    ...                 [1, 8]], dtype=tf.float16)
    >>> y_pred = tf.linalg.norm(a - b, axis=1)
    >>> y_pred
    <tf.Tensor: shape=(3,), dtype=float16, numpy=array([8.06 , 2.   , 4.473],
    dtype=float16)>

    <... Note: constants a & b have been used purely for
    example purposes and have no significant value ...>

    See: http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf

    Args:
      y_true: 1-D integer `Tensor` with shape `[batch_size]` of
        binary labels indicating positive vs negative pair.
      y_pred: 1-D float `Tensor` with shape `[batch_size]` of
        distances between two embedding matrices.
      margin: margin term in the loss definition.

    Returns:
      contrastive_loss: 1-D float `Tensor` with shape `[batch_size]`.
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.dtypes.cast(y_true, y_pred.dtype)
    return y_true * tf.math.square(y_pred) + (1.0 - y_true) * tf.math.square(
        tf.math.maximum(margin - y_pred, 0.0)
    )


from typing import Callable, Tuple

import numpy as np

from ._equal_error_thresholding_method import EqualErrorThresholding
from .distance_evaluation_method_base import DistanceBaseEvaluationMethod
from ..prediction_models.distance_prediction_model_base import DistanceBasePredictionModel
from ....util.dataset import Dataset
from ....util.utils import contrastive_make_pairs


class PairwiseAccuracy(DistanceBaseEvaluationMethod):

    def __init__(self,
                 output_dir: str = "evaluation_pairwise",
                 pair_function: Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]] =
                 contrastive_make_pairs):
        super().__init__(output_dir)
        self.pair_function = pair_function

    def evaluate(self, dataset: Dataset, prediction_model: DistanceBasePredictionModel, **kwargs):
        thresholding_method = EqualErrorThresholding(prediction_model, self.pair_function, self.output_dir)
        thresholding_method.compute_threshold(dataset.testset_matching_data, dataset.testset_matching_labels)



"""
Each Model instance implements some model. Each model needs:
1. Access to the Dataset object to know the shape,
2. Access to a Model Description to create the architecture,
"""
from typing import Callable

from useckit.v2.dataset import Dataset


class Model:
    def __init__(self, dataset, model_description):
        self.dataset: Dataset = dataset
        self.model_description: Callable = model_description
        self.model = self.build_model(self.model_description)

    def build_model(self, model_description):
        raise NotImplementedError("Given the model_description, the model must be built.")

    def fit(self, dataset):
        raise NotImplementedError("Given the dataset, the model must be fitted.")

    def predict(self, dataset):
        raise NotImplementedError("Given the dataset, the model must be used for prediction.")

    def persist(self, path_like):
        raise NotImplementedError("Given the path_like, the model must be persisted.")

    def load(self, path_like):
        raise NotImplementedError("Given the path_like, the model must be loaded.")



import numpy as np

from .tsc_evaluation_method_base import TSCBaseEvaluationMethod
from ..prediction_models.tsc_prediction_model_base import TSCBasePredictionModel
from ....evaluation.identification import IdentificationModel, perform_identification_evaluation
from ....util.dataset import Dataset


class TSCIdentificationModel(IdentificationModel):

    def __init__(self, dataset: Dataset, prediction_model: TSCBasePredictionModel):
        self.dataset = dataset
        self.prediction_model = prediction_model

    def identify(self, samples: np.ndarray) -> int:
        x_train, x_val, y_train, y_val, y_true, input_shape, nb_classes = \
            self.prediction_model.convert_dataset_to_legacy_values(self.dataset)
        x_test = samples
        _, _, _, y_test = self.dataset.view_one_hot_encoded_labels()
        y_pred = self.prediction_model.predict(x_test, y_true, x_train, y_train, y_test, return_df_metrics=False)
        return self.dataset.reverse_label_transform(y_pred)


class IdentificationOnly(TSCBaseEvaluationMethod):

    def __init__(self, output_dir: str = "evaluation_identification"):
        super().__init__(output_dir)

    def evaluate(self, dataset: Dataset, prediction_model: TSCBasePredictionModel, **kwargs):
        perform_identification_evaluation(TSCIdentificationModel(dataset, prediction_model), dataset,
                                          self.output_dir)



import sys

import numpy as np
from numpy import ndarray
from sklearn import metrics

from .anomaly_evaluation_method_base import AnomalyBaseEvaluationMethod
from ..prediction_models.anomaly_prediction_model_base import AnomalyBasePredictionModel
from ....util.dataset import Dataset
from ....util.plotting import plot_confusion_matrix


def _check_accuracy(threshold, predictions, assumed_label, true_labels):
    true = 0
    false = 0
    for prediction, true_label in zip(predictions, true_labels):
        if prediction <= threshold and assumed_label == true_label:
            true += 1
        else:
            false += 1
    return true / (true + false)


def _find_optimal_thersholds(thresholding_data: ndarray, thresholding_labels: ndarray,
                             prediction_model: AnomalyBasePredictionModel):
    thresh_predicts = prediction_model.predict(thresholding_data)
    model_thresholds = []
    for model_label, model_preds in enumerate(thresh_predicts):
        best_acc = 0.
        best_threshold = 0.
        for threshold_candidate in model_preds:
            acc = _check_accuracy(threshold_candidate, model_preds, model_label, thresholding_labels)
            if acc > best_acc:
                best_threshold = threshold_candidate
                best_acc = acc
        model_thresholds.append(best_threshold)
    return model_thresholds


def _predict_labels_from_thresholds(test_predicts: ndarray, model_thresholds: list):
    test_predicts = np.transpose(test_predicts)
    test_predict_labels = []
    for sample_id, sample_preds in enumerate(test_predicts):
        lowest_threshold_ratio = sys.float_info.max
        label = -1  # reject
        for model_label, pred in enumerate(sample_preds):
            if pred < model_thresholds[model_label] and pred / model_thresholds[model_label] < lowest_threshold_ratio:
                label = model_label
                lowest_threshold_ratio = pred / model_thresholds[model_label]
        test_predict_labels.append(label)

    return test_predict_labels


class LegacyIdentificationRejectIndividualThreshold(AnomalyBaseEvaluationMethod):

    def __init__(self, output_dir: str = "evaluation_identification_with_reject_legacy"):
        super().__init__(output_dir)

    def evaluate(self, dataset: Dataset, prediction_model: AnomalyBasePredictionModel, **kwargs):
        test_data, test_labels = dataset.testset_matching_data, dataset.testset_matching_labels
        thresholding_data, thresholding_labels = dataset.testset_enrollment_data, dataset.testset_enrollment_labels

        model_thresholds = _find_optimal_thersholds(thresholding_data, thresholding_labels, prediction_model)
        test_predicts = prediction_model.predict(test_data)

        number_models = test_predicts.shape[0]
        reject_mask = test_labels >= number_models  # users not seen during training
        test_labels = test_labels.copy()
        test_labels[reject_mask] = -1

        test_predict_labels = _predict_labels_from_thresholds(test_predicts, model_thresholds)
        contingency_table = metrics.confusion_matrix(test_labels, test_predict_labels)

        label_set = np.unique(np.concatenate((test_labels, test_predict_labels)))
        label_set = label_set[label_set != -1]
        label_names = dataset.reverse_label_transform(label_set)
        label_names = list(label_names)
        label_names.append("reject")
        plot_confusion_matrix(contingency_table, target_names=label_names, path=self.output_dir,
                              title='Contingency Table', filename='contingency_table.pdf', normalize=False)



from typing import Callable

import tensorflow.keras as keras

from useckit.paradigms.time_series_classification.prediction_models.instance_normalization_layer import \
    InstanceNormalization


def _dl4tsc_cnn(padding: str, input_layer: keras.layers.Input, nb_classes: int) -> (keras.models.Model, list[Callable]):
    conv1 = keras.layers.Conv1D(filters=6, kernel_size=7, padding=padding, activation='sigmoid')(input_layer)
    conv1 = keras.layers.AveragePooling1D(pool_size=3)(conv1)

    conv2 = keras.layers.Conv1D(filters=12, kernel_size=7, padding=padding, activation='sigmoid')(conv1)
    conv2 = keras.layers.AveragePooling1D(pool_size=3)(conv2)

    flatten_layer = keras.layers.Flatten()(conv2)

    output_layer = keras.layers.Dense(units=nb_classes, activation='sigmoid')(flatten_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='mean_squared_error', optimizer=keras.optimizers.Adam(),
                  metrics=['accuracy'])

    return model, []


# use with batch_size=16 and epochs=2000
def dl4tsc_cnn_padding_valid(input_layer: keras.layers.Input, nb_classes: int) -> (keras.models.Model, list[Callable]):
    padding = 'valid'
    return _dl4tsc_cnn(padding, input_layer, nb_classes)


# use with batch_size=16 and epochs=2000
def dl4tsc_cnn_padding_same(input_layer: keras.layers.Input, nb_classes: int) -> (keras.models.Model, list[Callable]):
    padding = 'same'
    return _dl4tsc_cnn(padding, input_layer, nb_classes)


# use with batch_size=16 and epochs=2000
def dl4tsc_fcn(input_layer: keras.layers.Input, nb_classes: int) -> (keras.models.Model, list[Callable]):
    conv1 = keras.layers.Conv1D(filters=128, kernel_size=8, padding='same')(input_layer)
    conv1 = keras.layers.BatchNormalization()(conv1)
    conv1 = keras.layers.Activation(activation='relu')(conv1)

    conv2 = keras.layers.Conv1D(filters=256, kernel_size=5, padding='same')(conv1)
    conv2 = keras.layers.BatchNormalization()(conv2)
    conv2 = keras.layers.Activation('relu')(conv2)

    conv3 = keras.layers.Conv1D(128, kernel_size=3, padding='same')(conv2)
    conv3 = keras.layers.BatchNormalization()(conv3)
    conv3 = keras.layers.Activation('relu')(conv3)

    gap_layer = keras.layers.GlobalAveragePooling1D()(conv3)

    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(gap_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(),
                  metrics=['accuracy'])

    reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50,
                                                  min_lr=0.0001)
    return model, [reduce_lr]


# use with batch_size=16 and epochs=5000
def dl4tsc_mlp(input_layer: keras.layers.Input, nb_classes: int) -> (keras.models.Model, list[Callable]):
    # flatten/reshape because when multivariate all should be on the same axis
    input_layer_flattened = keras.layers.Flatten()(input_layer)

    layer_1 = keras.layers.Dropout(0.1)(input_layer_flattened)
    layer_1 = keras.layers.Dense(500, activation='relu')(layer_1)

    layer_2 = keras.layers.Dropout(0.2)(layer_1)
    layer_2 = keras.layers.Dense(500, activation='relu')(layer_2)

    layer_3 = keras.layers.Dropout(0.2)(layer_2)
    layer_3 = keras.layers.Dense(500, activation='relu')(layer_3)

    output_layer = keras.layers.Dropout(0.3)(layer_3)
    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(output_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adadelta(),
                  metrics=['accuracy'])

    reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=200, min_lr=0.1)

    return model, [reduce_lr]


# use with batch_size=min(64, train_data/10) and epochs=1500
def dl4tsc_resnet(input_layer: keras.layers.Input, nb_classes: int) -> (keras.models.Model, list[Callable]):
    n_feature_maps = 64

    # BLOCK 1
    conv_x = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=8, padding='same')(input_layer)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    # expand channels for the sum
    shortcut_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=1, padding='same')(input_layer)
    shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

    output_block_1 = keras.layers.add([shortcut_y, conv_z])
    output_block_1 = keras.layers.Activation('relu')(output_block_1)

    # BLOCK 2

    conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8, padding='same')(output_block_1)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    # expand channels for the sum
    shortcut_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=1, padding='same')(output_block_1)
    shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

    output_block_2 = keras.layers.add([shortcut_y, conv_z])
    output_block_2 = keras.layers.Activation('relu')(output_block_2)

    # BLOCK 3

    conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8, padding='same')(output_block_2)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    # no need to expand channels because they are equal
    shortcut_y = keras.layers.BatchNormalization()(output_block_2)

    output_block_3 = keras.layers.add([shortcut_y, conv_z])
    output_block_3 = keras.layers.Activation('relu')(output_block_3)

    # FINAL

    gap_layer = keras.layers.GlobalAveragePooling1D()(output_block_3)

    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(gap_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(),
                  metrics=['accuracy'])

    reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50, min_lr=0.0001)

    return model, [reduce_lr]


# use with batch_size=12 and epochs=100
def dl4tsc_encoder(input_layer: keras.layers.Input, nb_classes: int) -> (keras.models.Model, list[Callable]):
    # conv block -1
    conv1 = keras.layers.Conv1D(filters=128, kernel_size=5, strides=1, padding='same')(input_layer)
    conv1 = InstanceNormalization()(conv1)
    conv1 = keras.layers.PReLU(shared_axes=[1])(conv1)
    conv1 = keras.layers.Dropout(rate=0.2)(conv1)
    conv1 = keras.layers.MaxPooling1D(pool_size=2)(conv1)
    # conv block -2
    conv2 = keras.layers.Conv1D(filters=256, kernel_size=11, strides=1, padding='same')(conv1)
    conv2 = InstanceNormalization()(conv2)
    conv2 = keras.layers.PReLU(shared_axes=[1])(conv2)
    conv2 = keras.layers.Dropout(rate=0.2)(conv2)
    conv2 = keras.layers.MaxPooling1D(pool_size=2)(conv2)
    # conv block -3
    conv3 = keras.layers.Conv1D(filters=512, kernel_size=21, strides=1, padding='same')(conv2)
    conv3 = InstanceNormalization()(conv3)
    conv3 = keras.layers.PReLU(shared_axes=[1])(conv3)
    conv3 = keras.layers.Dropout(rate=0.2)(conv3)
    # split for attention
    attention_data = keras.layers.Lambda(lambda x: x[:, :, :256])(conv3)
    attention_softmax = keras.layers.Lambda(lambda x: x[:, :, 256:])(conv3)
    # attention mechanism
    attention_softmax = keras.layers.Softmax()(attention_softmax)
    multiply_layer = keras.layers.Multiply()([attention_softmax, attention_data])
    # last layer
    dense_layer = keras.layers.Dense(units=256, activation='sigmoid')(multiply_layer)
    dense_layer = InstanceNormalization()(dense_layer)
    # output layer
    flatten_layer = keras.layers.Flatten()(dense_layer)
    output_layer = keras.layers.Dense(units=nb_classes, activation='softmax')(flatten_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(0.00001),
                  metrics=['accuracy'])

    return model, []





import os
import re
import shutil
import subprocess
import hashlib
import urllib.request
from glob import glob
from typing import List, Tuple, Optional

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tensorflow.keras.preprocessing.sequence import pad_sequences

from useckit.v2.dataset import WindowslicedDataset, Dataset


_DEF_DIRNAME = "chi-2021"
_FILE1 = "Data_Set_for_Understanding_User_Identification_in_Virtual_Reality_Through_Behavioral_Biometrics_and_the_Effect_of_Body_Normalization.7z.001"
_FILE2 = "Data_Set_for_Understanding_User_Identification_in_Virtual_Reality_Through_Behavioral_Biometrics_and_the_Effect_of_Body_Normalization.7z.002"
_URL1 = "https://hci.informatik.uni-due.de/fileadmin/fileupload/I-HCI/Paper/" + _FILE1
_URL2 = "https://hci.informatik.uni-due.de/fileadmin/fileupload/I-HCI/Paper/" + _FILE2
_SHA1 = "1235b224ba131c36fc70a29a2e381f8ec0f5d7aa51c6ad09e576068c2c1c8d7d"
_SHA2 = "3110f9603a07a0259eabff73c399ebb679880f28390d48ca6480623e2c79bda2"
_FILENAME_PATTERN = r'(?P<activity>\w+)_p(?P<pid>\d+)_(?P<condition>\w+)_session(?P<session>\d+)_repetition(?P<rep>\d+)\.csv'


def _download_dataset(dataset_dir: str, reinstall: bool = False) -> str:
    if reinstall and os.path.exists(dataset_dir):
        shutil.rmtree(dataset_dir)
    if not os.path.exists(dataset_dir):
        os.makedirs(dataset_dir, exist_ok=True)
        zip1 = os.path.join(dataset_dir, _FILE1)
        zip2 = os.path.join(dataset_dir, _FILE2)
        # Download
        urllib.request.urlretrieve(_URL1, zip1)
        urllib.request.urlretrieve(_URL2, zip2)
        # Hash check
        def sha256(path):
            with open(path, "rb") as f:
                return hashlib.sha256(f.read()).hexdigest()
        h1, h2 = sha256(zip1), sha256(zip2)
        print(f"sha256sum of {zip1} is {h1}")
        print(f"sha256sum of {zip2} is {h2}")
        assert h1 == _SHA1, "hashsum did not match expectation for .001"
        assert h2 == _SHA2, "hashsum did not match expectation for .002"
        # Extract using 7z
        try:
            subprocess.run(["7z", "x", zip1, "-o" + dataset_dir, "-y"], check=True)
        except FileNotFoundError as e:
            raise RuntimeError("7z (7-Zip) not found on PATH. Please install 7-Zip or add it to PATH.") from e
        # Cleanup parts
        os.remove(zip1)
        os.remove(zip2)
        # Sanity check
        csv_cnt = len(glob(os.path.join(dataset_dir, "*.csv")))
        assert csv_cnt == 3072, f"Expected 3072 csv files, found {csv_cnt}"
        print(f"Verified the extraction of {csv_cnt} csv files.")
    else:
        print(f"Using already installed dataset {dataset_dir} ...")
    return dataset_dir


def _load_single_csv(path: str) -> Tuple[pd.DataFrame, dict]:
    filename = os.path.basename(path)
    match = re.match(_FILENAME_PATTERN, filename)
    assert match, f"Did not find a match for file {filename}"
    meta = match.groupdict()
    df = pd.read_csv(path)
    # remove extra columns
    df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
    cols_to_drop = [
        "ParticipantID", "study_session", "repetition", "timestamp_ms", "phase",
        "HeightNormalization", "ArmLengthNormalization"
    ]
    cols_to_drop += [col for col in df.columns if
                     col.lower().startswith(("ball", "arrow", "rightcontrolleranchor",
                                             "leftcontrolleranchor", "scenario"))]
    df.drop(columns=cols_to_drop, inplace=True, errors='raise')
    return df, meta


def _preprocess_sample(df: pd.DataFrame) -> pd.DataFrame:
    from sklearn.preprocessing import MinMaxScaler
    # local vectors head -> virt. hands
    for lr in ["Left", "Right"]:
        for XYZ in ["X", "Y", "Z"]:
            df[f"{lr}VirtualHand_pos_{XYZ}"] -= df[f"CenterEyeAnchor_pos_{XYZ}"]
    df.drop(columns=[
        'CenterEyeAnchor_pos_X', 'CenterEyeAnchor_pos_Y', 'CenterEyeAnchor_pos_Z',
        'CenterEyeAnchor_euler_X', 'CenterEyeAnchor_euler_Y', 'CenterEyeAnchor_euler_Z'
    ], inplace=True)
    # make everything relative to the beginning
    for c in df.columns:
        df[c] -= df[c].iloc[0]
    # Min Max Scaling
    cols = df.columns
    arr = MinMaxScaler().fit_transform(df)
    return pd.DataFrame(arr, columns=cols)


def get_chi2021_windowsliced_dataset(
    window_size: int = 72,
    stride: int = 72//2,
    n_jobs: int = -1,
    dataset_path: Optional[str] = None,
    reinstall: bool = False,
    label_key: str = 'pid',
    shuffle_slices: bool = False,
) -> WindowslicedDataset:
    """
    Download (if necessary), preprocess and return the PyCHI 2021 dataset
    as a useckit WindowslicedDataset.

    - Samples are labeled by `label_key` parsed from filename (default: 'pid').
    - Only a single set (trainset) is constructed; validation/test can be split
      by the caller if needed.
    """
    # Resolve dataset directory
    base_dir = os.path.dirname(__file__)
    dataset_dir = dataset_path or os.path.join(base_dir, "resources", _DEF_DIRNAME)
    dataset_dir = _download_dataset(dataset_dir, reinstall=reinstall)

    file_paths = glob(os.path.join(dataset_dir, "*.csv"))

    def process_path(p: str):
        df, meta = _load_single_csv(p)
        df = _preprocess_sample(df)
        meta['pid'] = f"P{meta['pid']}"
        meta['session'] = int(meta['session'])
        return df, meta

    results = Parallel(n_jobs=n_jobs)(delayed(process_path)(p) for p in file_paths)

    x_train = [r[0].to_numpy() for r in results if r[1]['session'] == 1 and r[1]['condition'] == "WithoutNormalization"]
    y_train = [r[1]['pid'] for r in results if r[1]['session'] == 1 and r[1]['condition'] == "WithoutNormalization"]
    x_test = [r[0].to_numpy() for r in results if r[1]['session'] == 2 and r[1]['condition'] == "WithoutNormalization"]
    y_test = [r[1]['pid'] for r in results if r[1]['session'] == 2 and r[1]['condition'] == "WithoutNormalization"]

    ds = WindowslicedDataset(
        window_slicing_stride=stride,
        window_slicing_size=window_size,
        trainset_data=x_train,
        trainset_labels=y_train,
        testset_enrollment_data=x_train,
        testset_enrollment_labels=y_train,
        testset_matching_data=x_test,
        testset_matching_labels=y_test,
        normalization_check=False,
        shuffle_window_slices=shuffle_slices,
    )
    return ds


def get_chi2021_dataset(
    n_jobs: int = -1,
    dataset_path: Optional[str] = None,
    reinstall: bool = False,
    label_key: str = 'pid',
    filter: str = 'WithoutNormalization',
    padding: bool = False,
    feature_vector: bool = False,
    ws_enable: bool = False,
    ws_stride: int = 36,
    ws_size: int = 72,
) -> Dataset:
    """
    Download (if necessary), preprocess and return the CHI 2021 dataset
    as a regular Dataset (no window slicing), without any padding.

    - Samples are labeled by `label_key` parsed from filename (default: 'pid').
    - Train/test split follows the same rule as the windowsliced variant:
      session==1 -> enrollment (train), session==2 -> matching (test), both with condition=="WithoutNormalization".
    - Returns variable-length sequences as lists of np.ndarray (float32), one array per sample.
    """
    # Resolve dataset directory
    base_dir = os.path.dirname(__file__)
    dataset_dir = dataset_path or os.path.join(base_dir, "resources", _DEF_DIRNAME)
    dataset_dir = _download_dataset(dataset_dir, reinstall=reinstall)

    file_paths = glob(os.path.join(dataset_dir, "*.csv"))

    def process_path(p: str):
        df, meta = _load_single_csv(p)
        df = _preprocess_sample(df)
        meta['pid'] = f"P{meta['pid']}"
        meta['session'] = int(meta['session'])
        return df, meta

    results = Parallel(n_jobs=n_jobs)(delayed(process_path)(p) for p in file_paths)

    # Filter by condition and session
    def _sel(res, session):
        return [(r[0], r[1][label_key]) for r in res if r[1]['session'] == session and r[1]['condition'] == filter]

    train_pairs = _sel(results, session=1)
    test_pairs = _sel(results, session=2)

    if not train_pairs and not test_pairs:
        raise RuntimeError("No samples found for CHI2021 dataset with the given filters.")

    # Build lists of arrays (no padding)
    x_train = [df for df, _ in train_pairs]
    y_train = np.array([lbl for _, lbl in train_pairs], dtype=str)
    x_test = [df for df, _ in test_pairs]
    y_test = np.array([lbl for _, lbl in test_pairs], dtype=str)

    def q3(data):
        return float(np.percentile(data, 75))

    if padding:
        q3_len: int = int(q3([len(l) for l in x_train + x_test]))
        x_train = pad_sequences(x_train, maxlen=q3_len, dtype=np.float32, padding='pre')
        x_test = pad_sequences(x_test, maxlen=q3_len, dtype=np.float32, padding='pre')

    def calculate_featur_vector(df: pd.DataFrame):
        vector = []
        for c in df.columns:
            vector.append(df[c].min())
            vector.append(df[c].max())
            vector.append(df[c].mean())
            vector.append(df[c].std())
        return vector

    if feature_vector:
        x_train = [calculate_featur_vector(x) for x in x_train]
        x_test = [calculate_featur_vector(x) for x in x_test]

    if ws_enable:
        ds = WindowslicedDataset(
            window_slicing_stride=ws_stride,
            window_slicing_size=ws_size,
            trainset_data=x_train,
            trainset_labels=y_train,
            testset_enrollment_data=x_train,
            testset_enrollment_labels=y_train,
            testset_matching_data=x_test,
            testset_matching_labels=y_test,
            normalization_check=False,)
    else:
        ds = Dataset(
            trainset_data=x_train,
            trainset_labels=y_train,
            testset_enrollment_data=x_train,
            testset_enrollment_labels=y_train,
            testset_matching_data=x_test,
            testset_matching_labels=y_test,
            normalization_check=False,
        )

    return ds



from useckit.models.model_architectures import *
from useckit.util.dataset import Dataset
import keras
from .model_architectures import ModelArchitecture, Paradigm, mb_dl4tsc, mb_dl4tsc_lower

class ModelBuilder:
    def __init__(self, useckit_dataset: Dataset):
        self.data = useckit_dataset

    @staticmethod
    def check_valid(architecture, paradigm):
        if isinstance(architecture, str):
            architecture = ModelArchitecture.get_by_string(architecture)
        if isinstance(paradigm, str):
            paradigm = Paradigm.get_by_string(paradigm)

        # Check architecture and paradigm validity
        if not isinstance(architecture, ModelArchitecture):
            raise "architecture has to be a valid ModelArchitecture or name"
        if not isinstance(paradigm, Paradigm):
            raise "paradigm has to be a valid Paradigm or name"

        return architecture, paradigm

    def create_model(self, architecture: ModelArchitecture, paradigm: Paradigm,
                     pool_factor=None, kernel_size=None):
        """Central builder method to create a model.

        Expects a 
            1) architecture (e.g., 'resnet', 'mlp', ...), and 
            2) a paradigm (e.g., 'tsc', 'distancelearning', 'binaryverification', 'anomaly_detection').

        This method then returns a model that can be trained. The model that is returned has a distinct input and output.
        For input, the `uscekit_dataset` that is based via __init__ is used.
        For output, the respective returned model has a last layer and produces return-values that are subject to the paradigm.
        It is as follows:
        - for DistanceLearning, the model has a single output unit at the end which yields a distance obtained from the input-samples.
        - for BinaryVerification, the model has two output units that predict the probability if input falls into the trained class (= first neuron output),
            or any other class (= second neuron output). Therefore, [1, 0] would mean that the sample belongs to the trained class for 100%. The output is softmaxed.
        - for TimeSeriesClassification, the model returns a one-hot encoded vector from [0, ..., N] for N clsases. The output is softmaxed.
        """
        # Convert argument strings into enum
        architecture, paradigm = self.check_valid(architecture, paradigm)


        input_shape = self.data.trainset_data.shape[1:]
        output_size = len(self.data.get_unique_labels())
        
        input_layer, mid = mb_dl4tsc(architecture, input_shape, pool_factor, kernel_size)
        
        if paradigm == Paradigm.TSC:
            output_layer, args = mb_dl4tsc_lower(architecture, mid, output_size)
        if paradigm == Paradigm.BV:
            output_layer, args = mb_bv_lower(architecture, mid, output_size)
        if paradigm == Paradigm.AD:
            output_layer, args = mb_ad_lower(architecture, mid, output_size, input_shape)
        if paradigm == Paradigm.CL:
            output_layer, args = mb_dl4cl_lower(architecture, mid, output_size)
        if paradigm == Paradigm.TL:
            output_layer, args = mb_dl4tl_lower(architecture, mid, output_size) 
            
        
        model = keras.models.Model(inputs=input_layer, outputs=output_layer)

        model.compile(*args) # args = (optimizer, loss, metrics); this order is required
        
        return model



import numpy as np

from ._equal_error_thresholding_method import EqualErrorThresholding
from .distance_evaluation_method_base import DistanceBaseEvaluationMethod, BaseThresholdingMethod
from ..prediction_models.distance_prediction_model_base import DistanceBasePredictionModel
from ....evaluation.verification import VerificationModel, perform_verification_evaluation
from ....util.dataset import Dataset
from ....util.utils import contrastive_make_pairs


class DistanceVerificationModel(VerificationModel):

    def __init__(self,
                 distance_metric: DistanceBasePredictionModel,
                 enrollment_samples: np.ndarray,
                 enrollment_labels: np.ndarray,
                 rejection_threshold: float,
                 tradeoff_computation_speed_for_memory: bool):
        self.distance_metric = distance_metric
        self.enrollment_samples = enrollment_samples
        self.enrolment_labels = enrollment_labels
        self.rejection_threshold = rejection_threshold
        self.tradeoff_computation_speed_for_memory = tradeoff_computation_speed_for_memory

    def _create_distance_pairs_broadcasting(self, samples: np.ndarray, identity_claims: np.ndarray):
        sample_broadcast_array_shape = [len(samples) * len(self.enrollment_samples)]  # this is, in actuality, too big,
        # so the arrays need to be sliced down to size later
        for s in samples.shape[1:]:  # each other dimension needs to be equivalent to a single sample
            sample_broadcast_array_shape.append(s)
        sample_broadcast_array = np.zeros(shape=tuple(sample_broadcast_array_shape),
                                          dtype=self.enrollment_samples.dtype)
        enrolment_broadcast_array = np.zeros(shape=tuple(sample_broadcast_array_shape),
                                             dtype=self.enrollment_samples.dtype)
        sample_original_indices = np.zeros(
            shape=(len(samples) * len(self.enrollment_samples),))  # this will contain the original
        # indices of the broadcasted samples

        broadcast_index = 0
        for index, sample in enumerate(samples):
            i_claim = identity_claims[index]  # identity belonging to the one sample
            i_enrollment_samples = self.enrollment_samples[self.enrolment_labels == i_claim]  # all enrollment
            # samples of the claimed identity
            if len(i_enrollment_samples) <= 0:
                raise ValueError(f"Claimed identity {identity_claims[index]} was not part of the enrolled "
                                 f"identities {self.enrolment_labels}!")
            sample_broadcast = np.broadcast_to(sample, shape=i_enrollment_samples.shape)
            indices = np.ones(shape=(len(i_enrollment_samples),)) * index

            from_i = broadcast_index
            to_i = broadcast_index + len(i_enrollment_samples)

            sample_broadcast_array[from_i:to_i] = sample_broadcast
            enrolment_broadcast_array[from_i:to_i] = i_enrollment_samples
            sample_original_indices[from_i:to_i] = indices

            broadcast_index = to_i

        sample_broadcast_array = sample_broadcast_array[0:broadcast_index]
        enrolment_broadcast_array = enrolment_broadcast_array[0:broadcast_index]
        sample_original_indices = sample_original_indices[0:broadcast_index]  # broadcast_index points to the actual
        # length of this array
        distances = self.distance_metric.predict(sample_broadcast_array, enrolment_broadcast_array)

        return distances, sample_original_indices

    def _create_distance_pairs_iteration(self, samples: np.ndarray, identity_claims: np.ndarray):
        distances = np.zeros(
            shape=(len(samples) * len(self.enrollment_samples),))
        sample_original_indices = np.zeros(
            shape=(len(samples) * len(self.enrollment_samples),), dtype=int)  # this will contain the original
        # indices of the broadcasted samples. Both of these arrays are way too big and need to be sliced down to size
        # later

        broadcast_index = 0
        for index, sample in enumerate(samples):
            i_claim = identity_claims[index]  # identity claim belonging to the one sample
            i_enrollment_samples = self.enrollment_samples[self.enrolment_labels == i_claim]  # all enrollment
            # samples of the claimed identity
            if len(i_enrollment_samples) <= 0:
                raise ValueError(f"Claimed identity {identity_claims[index]} was not part of the enrolled "
                                 f"identities {self.enrolment_labels}!")
            sample_broadcast = np.broadcast_to(sample, shape=i_enrollment_samples.shape)
            indices = np.ones(shape=(len(i_enrollment_samples),)) * index

            from_i = broadcast_index
            to_i = broadcast_index + len(i_enrollment_samples)

            sample_distances = self.distance_metric.predict(sample_broadcast, i_enrollment_samples)
            distances[from_i:to_i] = sample_distances
            sample_original_indices[from_i:to_i] = indices

            broadcast_index = to_i

        distances = distances[0:broadcast_index]
        sample_original_indices = sample_original_indices[0:broadcast_index]  # broadcast_index points to the actual
        # length of this array
        return distances, sample_original_indices

    def verify(self, samples: np.ndarray, identity_claims: np.ndarray) -> bool:
        result = np.zeros(shape=(len(samples),))
        assert len(samples) == len(identity_claims)

        if self.tradeoff_computation_speed_for_memory:
            distances, sample_original_indices = self._create_distance_pairs_iteration(samples, identity_claims)
        else:
            distances, sample_original_indices = self._create_distance_pairs_broadcasting(samples, identity_claims)

        for index, sample in enumerate(samples):
            index_mask = sample_original_indices == index
            distances_to_enrollment = distances[index_mask]

            median_dist = np.median(distances_to_enrollment)
            result[index] = 1 if median_dist < self.rejection_threshold else 0
        return result == 1  # dirty transform to bool array


class Verification(DistanceBaseEvaluationMethod):

    def __init__(self,
                 output_dir: str = "evaluation_verification",
                 threshold_method: BaseThresholdingMethod = None,
                 tradeoff_computation_speed_for_memory: bool = True):
        super().__init__(output_dir)
        self.threshold_method = threshold_method
        self.tradeoff_computation_speed_for_memory = tradeoff_computation_speed_for_memory

    def evaluate(self, dataset: Dataset, prediction_model: DistanceBasePredictionModel, **kwargs):
        if self.threshold_method is None:
            self.threshold_method = EqualErrorThresholding(prediction_model, contrastive_make_pairs, self.output_dir)

        threshold = self.threshold_method.compute_threshold(dataset.testset_enrollment_data,
                                                            dataset.testset_enrollment_labels)
        verification_model = DistanceVerificationModel(prediction_model,
                                                       dataset.testset_enrollment_data,
                                                       dataset.testset_enrollment_labels,
                                                       threshold,
                                                       self.tradeoff_computation_speed_for_memory)
        perform_verification_evaluation(verification_model, dataset, self.output_dir)



import math
import math
import os

import numpy as np
import pandas as pd

from useckit.paradigms.anomaly_detection.evaluation_methods.anomaly_evaluation_method_base import \
    BaseThresholdingMethod
from useckit.paradigms.anomaly_detection.prediction_models.anomaly_prediction_model_base import \
    AnomalyBasePredictionModel


class FixedFalseNegativeThresholdingMethod(BaseThresholdingMethod):

    def __init__(self, anomaly_detection_model: AnomalyBasePredictionModel, output_dir: str,
                 target_false_negative_rate=0.05):
        self.target_false_negative_rate = target_false_negative_rate
        self.anomaly_detection_model = anomaly_detection_model
        self.output_dir = output_dir

    def compute_thresholds(self, enrollment_data: np.ndarray, enrollment_labels: np.ndarray) -> [float]:
        result = dict()
        prediction, labels = self.anomaly_detection_model.predict(enrollment_data)

        columns = [f'{i}\'th model prediction' for i in list(range(len(prediction)))]
        pred_df = pd.DataFrame(prediction.transpose(), columns=columns)
        pred_df.to_csv(os.path.join(self.output_dir, 'enrollment_data_raw_predictions.csv'))

        for user_index, anomaly_predictions in zip(labels, prediction):
            same_user_predictions = anomaly_predictions[enrollment_labels == user_index]
            same_user_predictions = np.sort(same_user_predictions)
            threshold_cutoff_index = len(same_user_predictions) * (1 - self.target_false_negative_rate)
            threshold_cutoff_index = math.floor(threshold_cutoff_index)
            result[user_index] = same_user_predictions[threshold_cutoff_index]

        result_copy = dict()
        for key, value in result.items():
            result_copy[str(key)] = [value]  # need to cast keys to str for pandas...
        result_df = pd.DataFrame(result_copy)
        result_df.to_csv(os.path.join(self.output_dir, 'model_thresholds.csv'))
        return result



from typing import Callable, Union, Optional

import tensorflow as tf


def is_tensor_or_variable(x):
    return tf.is_tensor(x) or isinstance(x, tf.Variable)


def _masked_maximum(data, mask, dim=1):
    """Computes the axis wise maximum over chosen elements.

    Args:
      data: 2-D float `Tensor` of shape `[n, m]`.
      mask: 2-D Boolean `Tensor` of shape `[n, m]`.
      dim: The dimension over which to compute the maximum.

    Returns:
      masked_maximums: N-D `Tensor`.
        The maximized dimension is of size 1 after the operation.
    """
    axis_minimums = tf.math.reduce_min(data, dim, keepdims=True)
    masked_maximums = (
        tf.math.reduce_max(
            tf.math.multiply(data - axis_minimums, mask), dim, keepdims=True
        )
        + axis_minimums
    )
    return masked_maximums


def _masked_minimum(data, mask, dim=1):
    """Computes the axis wise minimum over chosen elements.

    Args:
      data: 2-D float `Tensor` of shape `[n, m]`.
      mask: 2-D Boolean `Tensor` of shape `[n, m]`.
      dim: The dimension over which to compute the minimum.

    Returns:
      masked_minimums: N-D `Tensor`.
        The minimized dimension is of size 1 after the operation.
    """
    axis_maximums = tf.math.reduce_max(data, dim, keepdims=True)
    masked_minimums = (
        tf.math.reduce_min(
            tf.math.multiply(data - axis_maximums, mask), dim, keepdims=True
        )
        + axis_maximums
    )
    return masked_minimums


def pairwise_distance(feature, squared: bool = False):
    """Computes the pairwise distance matrix with numerical stability.

    output[i, j] = || feature[i, :] - feature[j, :] ||_2

    Args:
      feature: 2-D Tensor of size `[number of data, feature dimension]`.
      squared: Boolean, whether or not to square the pairwise distances.

    Returns:
      pairwise_distances: 2-D Tensor of size `[number of data, number of data]`.
    """
    pairwise_distances_squared = tf.math.add(
        tf.math.reduce_sum(tf.math.square(feature), axis=[1], keepdims=True),
        tf.math.reduce_sum(
            tf.math.square(tf.transpose(feature)), axis=[0], keepdims=True
        ),
    ) - 2.0 * tf.matmul(feature, tf.transpose(feature))

    # Deal with numerical inaccuracies. Set small negatives to zero.
    pairwise_distances_squared = tf.math.maximum(pairwise_distances_squared, 0.0)
    # Get the mask where the zero distances are at.
    error_mask = tf.math.less_equal(pairwise_distances_squared, 0.0)

    # Optionally take the sqrt.
    if squared:
        pairwise_distances = pairwise_distances_squared
    else:
        pairwise_distances = tf.math.sqrt(
            pairwise_distances_squared
            + tf.cast(error_mask, dtype=tf.dtypes.float32) * 1e-16
        )

    # Undo conditionally adding 1e-16.
    pairwise_distances = tf.math.multiply(
        pairwise_distances,
        tf.cast(tf.math.logical_not(error_mask), dtype=tf.dtypes.float32),
    )

    num_data = tf.shape(feature)[0]
    # Explicitly set diagonals to zero.
    mask_offdiagonals = tf.ones_like(pairwise_distances) - tf.linalg.diag(
        tf.ones([num_data])
    )
    pairwise_distances = tf.math.multiply(pairwise_distances, mask_offdiagonals)
    return pairwise_distances


def angular_distance(feature):
    """Computes the angular distance matrix.

    output[i, j] = 1 - cosine_similarity(feature[i, :], feature[j, :])

    Args:
      feature: 2-D Tensor of size `[number of data, feature dimension]`.

    Returns:
      angular_distances: 2-D Tensor of size `[number of data, number of data]`.
    """
    # normalize input
    feature = tf.math.l2_normalize(feature, axis=1)

    # create adjaceny matrix of cosine similarity
    angular_distances = 1 - tf.matmul(feature, feature, transpose_b=True)

    # ensure all distances > 1e-16
    angular_distances = tf.maximum(angular_distances, 0.0)

    return angular_distances


def triplet_hard_loss(
    y_true,
    y_pred,
    margin = 1.0,
    soft: bool = False,
    distance_metric: Union[str, Callable] = "L2",
) -> tf.Tensor:
    r"""Computes the triplet loss with hard negative and hard positive mining.

    Usage:

    >>> y_true = tf.convert_to_tensor([0, 0])
    >>> y_pred = tf.convert_to_tensor([[0.0, 1.0], [1.0, 0.0]])
    >>> tfa.losses.triplet_hard_loss(y_true, y_pred, distance_metric="L2")
    <tf.Tensor: shape=(), dtype=float32, numpy=1.0>

    >>> # Calling with callable `distance_metric`
    >>> distance_metric = lambda x: tf.linalg.matmul(x, x, transpose_b=True)
    >>> tfa.losses.triplet_hard_loss(y_true, y_pred, distance_metric=distance_metric)
    <tf.Tensor: shape=(), dtype=float32, numpy=0.0>

    Args:
      y_true: 1-D integer `Tensor` with shape `[batch_size]` of
        multiclass integer labels.
      y_pred: 2-D float `Tensor` of embedding vectors. Embeddings should
        be l2 normalized.
      margin: Float, margin term in the loss definition.
      soft: Boolean, if set, use the soft margin version.
      distance_metric: `str` or a `Callable` that determines distance metric.
        Valid strings are "L2" for l2-norm distance,
        "squared-L2" for squared l2-norm distance,
        and "angular" for cosine similarity.

        A `Callable` should take a batch of embeddings as input and
        return the pairwise distance matrix.

    Returns:
      triplet_loss: float scalar with dtype of `y_pred`.
    """
    labels = tf.convert_to_tensor(y_true, name="labels")
    embeddings = tf.convert_to_tensor(y_pred, name="embeddings")

    convert_to_float32 = (
        embeddings.dtype == tf.dtypes.float16 or embeddings.dtype == tf.dtypes.bfloat16
    )
    precise_embeddings = (
        tf.cast(embeddings, tf.dtypes.float32) if convert_to_float32 else embeddings
    )

    # Reshape label tensor to [batch_size, 1].
    lshape = tf.shape(labels)
    labels = tf.reshape(labels, [lshape[0], 1])

    # Build pairwise squared distance matrix.
    if distance_metric == "L2":
        pdist_matrix = pairwise_distance(
            precise_embeddings, squared=False
        )

    elif distance_metric == "squared-L2":
        pdist_matrix = pairwise_distance(
            precise_embeddings, squared=True
        )

    elif distance_metric == "angular":
        pdist_matrix = angular_distance(precise_embeddings)

    else:
        pdist_matrix = distance_metric(precise_embeddings)

    # Build pairwise binary adjacency matrix.
    adjacency = tf.math.equal(labels, tf.transpose(labels))
    # Invert so we can select negatives only.
    adjacency_not = tf.math.logical_not(adjacency)

    adjacency_not = tf.cast(adjacency_not, dtype=tf.dtypes.float32)
    # hard negatives: smallest D_an.
    hard_negatives = _masked_minimum(pdist_matrix, adjacency_not)

    batch_size = tf.size(labels)

    adjacency = tf.cast(adjacency, dtype=tf.dtypes.float32)

    mask_positives = tf.cast(adjacency, dtype=tf.dtypes.float32) - tf.linalg.diag(
        tf.ones([batch_size])
    )

    # hard positives: largest D_ap.
    hard_positives = _masked_maximum(pdist_matrix, mask_positives)

    if soft:
        triplet_loss = tf.math.log1p(tf.math.exp(hard_positives - hard_negatives))
    else:
        triplet_loss = tf.maximum(hard_positives - hard_negatives + margin, 0.0)

    # Get final mean triplet loss
    triplet_loss = tf.reduce_mean(triplet_loss)

    if convert_to_float32:
        return tf.cast(triplet_loss, embeddings.dtype)
    else:
        return triplet_loss


def triplet_semihard_loss(
    y_true,
    y_pred,
    margin=1.0,
    distance_metric: Union[str, Callable] = "L2",
) -> tf.Tensor:
    r"""Computes the triplet loss with semi-hard negative mining.

    Usage:

    >>> y_true = tf.convert_to_tensor([0, 0])
    >>> y_pred = tf.convert_to_tensor([[0.0, 1.0], [1.0, 0.0]])
    >>> tfa.losses.triplet_semihard_loss(y_true, y_pred, distance_metric="L2")
    <tf.Tensor: shape=(), dtype=float32, numpy=2.4142137>

    >>> # Calling with callable `distance_metric`
    >>> distance_metric = lambda x: tf.linalg.matmul(x, x, transpose_b=True)
    >>> tfa.losses.triplet_semihard_loss(y_true, y_pred, distance_metric=distance_metric)
    <tf.Tensor: shape=(), dtype=float32, numpy=1.0>

    Args:
      y_true: 1-D integer `Tensor` with shape `[batch_size]` of
        multiclass integer labels.
      y_pred: 2-D float `Tensor` of embedding vectors. Embeddings should
        be l2 normalized.
      margin: Float, margin term in the loss definition.
      distance_metric: `str` or a `Callable` that determines distance metric.
        Valid strings are "L2" for l2-norm distance,
        "squared-L2" for squared l2-norm distance,
        and "angular" for cosine similarity.

        A `Callable` should take a batch of embeddings as input and
        return the pairwise distance matrix.

    Returns:
      triplet_loss: float scalar with dtype of `y_pred`.
    """
    labels = tf.convert_to_tensor(y_true, name="labels")
    embeddings = tf.convert_to_tensor(y_pred, name="embeddings")

    convert_to_float32 = (
        embeddings.dtype == tf.dtypes.float16 or embeddings.dtype == tf.dtypes.bfloat16
    )
    precise_embeddings = (
        tf.cast(embeddings, tf.dtypes.float32) if convert_to_float32 else embeddings
    )

    # Reshape label tensor to [batch_size, 1].
    lshape = tf.shape(labels)
    labels = tf.reshape(labels, [lshape[0], 1])

    # Build pairwise squared distance matrix

    if distance_metric == "L2":
        pdist_matrix = pairwise_distance(
            precise_embeddings, squared=False
        )

    elif distance_metric == "squared-L2":
        pdist_matrix = pairwise_distance(
            precise_embeddings, squared=True
        )

    elif distance_metric == "angular":
        pdist_matrix = angular_distance(precise_embeddings)

    else:
        pdist_matrix = distance_metric(precise_embeddings)

    # Build pairwise binary adjacency matrix.
    adjacency = tf.math.equal(labels, tf.transpose(labels))
    # Invert so we can select negatives only.
    adjacency_not = tf.math.logical_not(adjacency)

    batch_size = tf.size(labels)

    # Compute the mask.
    pdist_matrix_tile = tf.tile(pdist_matrix, [batch_size, 1])
    mask = tf.math.logical_and(
        tf.tile(adjacency_not, [batch_size, 1]),
        tf.math.greater(
            pdist_matrix_tile, tf.reshape(tf.transpose(pdist_matrix), [-1, 1])
        ),
    )
    mask_final = tf.reshape(
        tf.math.greater(
            tf.math.reduce_sum(
                tf.cast(mask, dtype=tf.dtypes.float32), 1, keepdims=True
            ),
            0.0,
        ),
        [batch_size, batch_size],
    )
    mask_final = tf.transpose(mask_final)

    adjacency_not = tf.cast(adjacency_not, dtype=tf.dtypes.float32)
    mask = tf.cast(mask, dtype=tf.dtypes.float32)

    # negatives_outside: smallest D_an where D_an > D_ap.
    negatives_outside = tf.reshape(
        _masked_minimum(pdist_matrix_tile, mask), [batch_size, batch_size]
    )
    negatives_outside = tf.transpose(negatives_outside)

    # negatives_inside: largest D_an.
    negatives_inside = tf.tile(
        _masked_maximum(pdist_matrix, adjacency_not), [1, batch_size]
    )
    semi_hard_negatives = tf.where(mask_final, negatives_outside, negatives_inside)

    loss_mat = tf.math.add(margin, pdist_matrix - semi_hard_negatives)

    mask_positives = tf.cast(adjacency, dtype=tf.dtypes.float32) - tf.linalg.diag(
        tf.ones([batch_size])
    )

    # In lifted-struct, the authors multiply 0.5 for upper triangular
    #   in semihard, they take all positive pairs except the diagonal.
    num_positives = tf.math.reduce_sum(mask_positives)

    triplet_loss = tf.math.truediv(
        tf.math.reduce_sum(
            tf.math.maximum(tf.math.multiply(loss_mat, mask_positives), 0.0)
        ),
        num_positives,
    )

    if convert_to_float32:
        return tf.cast(triplet_loss, embeddings.dtype)
    else:
        return triplet_loss

class LossFunctionWrapper(tf.keras.losses.Loss):
    """Wraps a loss function in the `Loss` class."""

    def __init__(
        self, fn, reduction=tf.keras.losses.Reduction.AUTO, name=None, **kwargs
    ):
        """Initializes `LossFunctionWrapper` class.

        Args:
          fn: The loss function to wrap, with signature `fn(y_true, y_pred,
            **kwargs)`.
          reduction: (Optional) Type of `tf.keras.losses.Reduction` to apply to
            loss. Default value is `AUTO`. `AUTO` indicates that the reduction
            option will be determined by the usage context. For almost all cases
            this defaults to `SUM_OVER_BATCH_SIZE`. When used with
            `tf.distribute.Strategy`, outside of built-in training loops such as
            `tf.keras` `compile` and `fit`, using `AUTO` or `SUM_OVER_BATCH_SIZE`
            will raise an error. Please see this custom training [tutorial](
              https://www.tensorflow.org/tutorials/distribute/custom_training)
            for more details.
          name: (Optional) name for the loss.
          **kwargs: The keyword arguments that are passed on to `fn`.
        """
        super().__init__(reduction=reduction, name=name)
        self.fn = fn
        self._fn_kwargs = kwargs

    def call(self, y_true, y_pred):
        """Invokes the `LossFunctionWrapper` instance.

        Args:
          y_true: Ground truth values.
          y_pred: The predicted values.

        Returns:
          Loss values per sample.
        """
        return self.fn(y_true, y_pred, **self._fn_kwargs)

    def get_config(self):
        config = {}
        for k, v in iter(self._fn_kwargs.items()):
            config[k] = tf.keras.backend.eval(v) if is_tensor_or_variable(v) else v
        base_config = super().get_config()
        return {**base_config, **config}


def normalize_data_format(value):
    if value is None:
        value = tf.keras.backend.image_data_format()
    data_format = value.lower()
    if data_format not in {"channels_first", "channels_last"}:
        raise ValueError(
            "The `data_format` argument must be one of "
            '"channels_first", "channels_last". Received: ' + str(value)
        )
    return data_format


def normalize_tuple(value, n, name):
    """Transforms an integer or iterable of integers into an integer tuple.

    A copy of tensorflow.python.keras.util.

    Args:
      value: The value to validate and convert. Could an int, or any iterable
        of ints.
      n: The size of the tuple to be returned.
      name: The name of the argument being validated, e.g. "strides" or
        "kernel_size". This is only used to format error messages.

    Returns:
      A tuple of n integers.

    Raises:
      ValueError: If something else than an int/long or iterable thereof was
        passed.
    """
    if isinstance(value, int):
        return (value,) * n
    else:
        try:
            value_tuple = tuple(value)
        except TypeError:
            raise TypeError(
                "The `"
                + name
                + "` argument must be a tuple of "
                + str(n)
                + " integers. Received: "
                + str(value)
            )
        if len(value_tuple) != n:
            raise ValueError(
                "The `"
                + name
                + "` argument must be a tuple of "
                + str(n)
                + " integers. Received: "
                + str(value)
            )
        for single_value in value_tuple:
            try:
                int(single_value)
            except (ValueError, TypeError):
                raise ValueError(
                    "The `"
                    + name
                    + "` argument must be a tuple of "
                    + str(n)
                    + " integers. Received: "
                    + str(value)
                    + " "
                    "including element "
                    + str(single_value)
                    + " of type"
                    + " "
                    + str(type(single_value))
                )
        return value_tuple



class TripletSemiHardLoss(LossFunctionWrapper):
    """Computes the triplet loss with semi-hard negative mining.

    The loss encourages the positive distances (between a pair of embeddings
    with the same labels) to be smaller than the minimum negative distance
    among which are at least greater than the positive distance plus the
    margin constant (called semi-hard negative) in the mini-batch.
    If no such negative exists, uses the largest negative distance instead.
    See: https://arxiv.org/abs/1503.03832.

    We expect labels `y_true` to be provided as 1-D integer `Tensor` with shape
    `[batch_size]` of multi-class integer labels. And embeddings `y_pred` must be
    2-D float `Tensor` of l2 normalized embedding vectors.

    Args:
      margin: Float, margin term in the loss definition. Default value is 1.0.
      name: Optional name for the op.
    """

    def __init__(
        self,
        margin = 1.0,
        distance_metric: Union[str, Callable] = "L2",
        name: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            triplet_semihard_loss,
            name=name,
            reduction=tf.keras.losses.Reduction.NONE,
            margin=margin,
            distance_metric=distance_metric,
        )


import json
from collections import Counter
from pathlib import Path

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from sklearn.metrics import auc

from useckit import Dataset
from useckit.paradigms.binary_verification.evaluation_methods.binary_verification_evaluation_method_base import \
    VerificationBaseEvaluationMethod
from useckit.paradigms.binary_verification.prediction_models.verification_prediction_model_base import \
    VerificationBasePredictionModel


class BinaryScoringVerification(VerificationBaseEvaluationMethod):
    def evaluate(self, dataset: Dataset, prediction_model: VerificationBasePredictionModel, **kwargs):
        # The `prediction_model` contains fitted classifiers.

        # Extract multiclass_strategy from kwargs or set default to 'ovr'
        multiclass_strategy = kwargs["multiclass_classification_strategy"].lower()
        if multiclass_strategy not in ["ovr", "ovo"]:
            raise ValueError(f'multiclass_strategy must be either "ovr" or "ovo", but '
                             f'{multiclass_strategy} was provided.')
        self.multiclass_strategy = multiclass_strategy

        # Unpack prediction_model
        classifs = prediction_model.classifs
        self.output_dir = Path(prediction_model.output_dir).parent / "evaluation"
        self.verbose = prediction_model.verbose

        # Unpack dataset
        reject_label = dataset.reject_label
        matching_data = dataset.testset_matching_data
        matching_labels = dataset.testset_matching_labels

        # Prepare result storage
        if multiclass_strategy == "ovr":
            graph = self._run_ovr_with_matchingdata(classifs, matching_data, matching_labels,
                                                    getattr(dataset, 'testset_matching_slicedsample_origin', None))
        elif multiclass_strategy == "ovo":
            graph = self._run_ovo_with_matchingdata(classifs, matching_data, matching_labels,
                                                    getattr(dataset, 'testset_matching_slicedsample_origin', None))

        proba_metrics: pd.DataFrame = self._calculate_proba_based_predictions(graph,
                                                                              reverse_decision=kwargs.get("reverse_decision", False))

        proba_metrics.to_csv(Path(self.output_dir) / 'threshold-metrics.tsv', sep='\t', index=False)

        self.calc_and_plot_eer(proba_metrics)
        self.calc_and_plot_auc(proba_metrics)

    def _run_ovr_with_matchingdata(self, classifs, matching_data, matching_labels, slicedsample_origin=None):
        results = []

        # Test matching data for all classifiers
        for key in classifs.keys():
            classif = classifs[key]

            y_scores = classif.predict_proba(matching_data)[:, 0]
            y_true = (matching_labels == key).astype(int)

            results.append({'clf_class one label': key,
                            'y_scores': y_scores,
                            'y_true': y_true,
                            'matching_labels': matching_labels,
                            'slicedsample_origin': slicedsample_origin,  # can be None
                            'method': 'ovr'})

        return results

    def _run_ovo_with_matchingdata(self, classifs, matching_data, matching_labels, slicedsample_origin=None):
        results = []

        for key1 in classifs.keys():
            for key2 in classifs[key1].keys():
                classif = classifs[key1][key2]

                # Select data where the matching label is either label_one or label_other
                data_mask = (matching_labels == key1) | (matching_labels == key2)
                selected_data = matching_data[data_mask]
                selected_labels = matching_labels[data_mask]
                selected_slice_samples = slicedsample_origin[data_mask] if slicedsample_origin is not None else None

                # Compute probability scores for class corresponding to label_one
                y_scores = classif.predict_proba(selected_data)[:, 0]

                # Create true labels array: 1 where the label matches label_one, otherwise 0
                y_true = (selected_labels == key1).astype(int)

                # runtime validation
                if slicedsample_origin is not None:
                    assert len(y_scores) == len(y_true) == len(selected_labels) == len(selected_slice_samples), \
                        (f"The lengths of `y_scores` (len={len(y_scores)}), `y_true` (len={len(y_true)}), "
                         f"`selected_labels` (len={len(selected_labels)}), and `selected_slice_samples` "
                         f"(len={len(selected_slice_samples)}) must match, but they are not equal.")
                else:
                    assert len(y_scores) == len(y_true) == len(selected_labels), \
                        (f"The lengths of `y_scores` (len={len(y_scores)}), `y_true` (len={len(y_true)}), "
                         f"and `selected_labels` (len={len(selected_labels)}), must match, but they are not equal.")

                # Append results
                results.append({
                    'clf_class one label': key1,
                    'clf_class other label': key2,
                    'y_scores': y_scores,
                    'y_true': y_true,
                    'matching_labels': selected_labels,
                    'slicedsample_origin': selected_slice_samples,  # can be None when window slicing is not enabled
                    'method': 'ovo'
                })

        return results

    def _calculate_proba_based_predictions(self, verification_graph: list[dict], reverse_decision: bool = False) -> pd.DataFrame:
        all_possible_thresholds = np.unique(np.concatenate([r['y_scores'] for r in verification_graph]))
        epsilon = np.finfo(float).eps

        # add edge cases with the smallest possible adjustments
        min_score = min(all_possible_thresholds) - epsilon
        max_score = max(all_possible_thresholds) + epsilon

        # contains all possible thresholds
        all_possible_thresholds = np.concatenate([[min_score], all_possible_thresholds, [max_score]])

        # array to store results
        metric_results = []

        # bested loops for all combinations
        for i, threshold in enumerate(all_possible_thresholds):
            for edge in verification_graph:  # for every identity
                current_class_label_gt = edge['clf_class one label']  # current ground truth label
                slicedsample_origin = edge.get('slicedsample_origin')  # can be "None" if window slicing is not enabled

                window_slicing_is_enabled: bool = slicedsample_origin is not None

                if window_slicing_is_enabled:
                    # get unique samples if windowslicing is enabled so that we can initialize the voting_df
                    unique_samples: np.ndarray = np.unique(slicedsample_origin)
                else:
                    # if no window_slicing; create mask for compatibility
                    unique_samples = np.arange(len(edge['y_scores']))

                # initialization of one row per unique_sample with all metrics set to zero
                # index is: (sample_origin, threshold)
                voting_df = pd.DataFrame({'sample_origin': unique_samples,
                                          'threshold': [threshold]*len(unique_samples),
                                          'tp': [0]*len(unique_samples),
                                          'fp': [0]*len(unique_samples),
                                          'tn': [0]*len(unique_samples),
                                          'fn': [0]*len(unique_samples)}).set_index(["sample_origin", "threshold"])

                for usample in unique_samples:
                    if window_slicing_is_enabled:
                         # get all predictions for every unique sample sample when window slicing
                        get_mask: np.ndarray = edge['slicedsample_origin'] == usample
                    else:
                        # window slicing is not enabled
                        get_mask: np.ndarray = unique_samples == usample

                    # unpack edge for current unique sample
                    usample_scores_array: np.ndarray = edge['y_scores'][get_mask]  # array of prediction scores from clf
                    usample_gt_array: np.ndarray = edge['y_true'][get_mask]  # array of truth predictions (1=accept, 0=reject)
                    usample_matching_label_array: np.ndarray = edge['matching_labels'][get_mask]  # array of gt class labels

                    sample_ground_truth: np.ndarray = np.unique(usample_gt_array)
                    sample_matching_label: np.ndarray = np.unique(usample_matching_label_array)

                    # assert that no mix up happened between ground truths and matching labels
                    assert len(sample_ground_truth) == 1, (f"Expected `sample_ground_truth` to have only one "
                                                           f"element, but found {len(sample_ground_truth)} "
                                                           f"elements: {sample_ground_truth}")
                    assert len(sample_matching_label) == 1, (f"Expected `sample_matching_label` to have only one "
                                                             f"element, but found {len(sample_matching_label)} "
                                                             f"elements: {sample_matching_label}")

                    sample_ground_truth: int = sample_ground_truth[0]
                    sample_matching_label: int = sample_matching_label[0]

                    # majority_vote_values contains bools denoting whether sample should be accepted or rejected
                    if not reverse_decision:
                        majority_vote_values, majority_vote_counts = Counter(usample_scores_array < threshold).most_common()[0]
                    else:
                        majority_vote_values, majority_vote_counts = Counter(usample_scores_array >= threshold).most_common()[0]


                    accept_sample: bool = majority_vote_values
                    matching_label_equals_current_class: bool = current_class_label_gt == sample_matching_label

                    if accept_sample and matching_label_equals_current_class:
                        voting_df.loc[(usample, threshold), "tp"] += 1
                    elif accept_sample and not matching_label_equals_current_class:
                        voting_df.loc[(usample, threshold), "fp"] += 1
                    elif not accept_sample and matching_label_equals_current_class:
                        voting_df.loc[(usample, threshold), "fn"] += 1
                    elif not accept_sample and not matching_label_equals_current_class:
                        voting_df.loc[(usample, threshold), "tn"] += 1

                # store into list
                metric_results.append(voting_df)

        # concatenate voting_dfs for various thresholds and samples
        metric_df = pd.concat(metric_results).reset_index()

        # sum up how many TP, FP, TN, and FN exist per. sample_origin loses its meaning upon grouping by threshold.
        metric_df = metric_df.groupby("threshold").sum().drop(columns=["sample_origin"])

        assert len(np.unique(metric_df.sum(axis=1))) == 1, (f'There is an offset in sample calculation. '
                                                            f'Every row is suspected to have an equal amount of '
                                                            f'predictions per threshold, but instead more than one '
                                                            f'value was found: {np.unique(metric_df.sum(axis=1))}')

        # add additional metrics
        metric_df['tpr'] = metric_df.apply(lambda row: self.calculate_tpr(row['tp'], row['fn']), axis=1)
        metric_df['fpr'] = metric_df.apply(lambda row: self.calculate_fpr(row['fp'], row['tn']), axis=1)
        metric_df['tnr'] = metric_df.apply(lambda row: self.calculate_tnr(row['tn'], row['fp']), axis=1)
        metric_df['fnr'] = metric_df.apply(lambda row: self.calculate_fnr(row['fn'], row['tp']), axis=1)
        metric_df['acc'] = metric_df.apply(lambda row: self.calculate_acc(row['tp'], row['tn'], row['fp'], row['fn']),
                                           axis=1)
        metric_df['bacc'] = metric_df.apply(lambda row: self.calculate_bacc(row['tp'], row['tn'], row['fp'], row['fn']),
                                           axis=1)

        return metric_df.reset_index()

    def calc_and_plot_eer(self, df: pd.DataFrame):
        # Calculate the point where FPR and FNR are closest (EER)
        df['distance fpr to fnr'] = np.abs(df['fpr'] - df['fnr'])
        eer_threshold = df.loc[df['distance fpr to fnr'].idxmin(), 'threshold']
        eer_value = df.loc[df['distance fpr to fnr'].idxmin(), 'fpr']

        # Plotting
        plt.figure(figsize=(10, 5))
        plt.plot(df['threshold'], df['fpr'], label='False Positive Rate (FPR / FAR)')
        plt.plot(df['threshold'], df['fnr'], label='False Negative Rate (FNR / FRR)')
        plt.axvline(x=eer_threshold, color='r', linestyle='--', label=f'EER {eer_value:.4f} at threshold {eer_threshold:.4f}')
        plt.title('Equal Error Rate (EER) Analysis')
        plt.xlabel('Threshold')
        plt.ylabel('Error Rate')
        plt.legend()
        plt.grid(True)
        plt.savefig(Path(self.output_dir) / 'eer.pdf', bbox_inches="tight")

        if self.verbose:
            print(f"The Equal Error Rate (EER) is approximately {eer_value:.2f} at threshold {eer_threshold:.5f}")

        # Write data to a JSON file
        with open(Path(self.output_dir) / 'eer.json', 'w') as file:
            json.dump({"eer_threshold": eer_threshold, "eer_value": eer_value}, file, indent=4)

    def calc_and_plot_auc(self, df: pd.DataFrame):
        # Calculate the ROC curve and AUC
        df_sorted = df.sort_values(by="fpr")
        roc_auc = auc(x=df_sorted["fpr"].values, y=df_sorted["tpr"].values)

        # Plotting
        plt.figure(figsize=(6, 5))
        plt.plot(df.fpr, df.tpr, color='darkorange', lw=2, label=f'ROC curve (AUC = {roc_auc:.2f})')
        plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--', label=f'Random Guess (AUC = 0.5)')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Receiver Operating Characteristic (ROC)')
        plt.legend(loc="lower right")
        plt.grid(True)
        plt.savefig(Path(self.output_dir) / 'roc_curve.pdf', bbox_inches="tight")

        if self.verbose:
            print(f"The Area Under the Curve (AUC) is {roc_auc:.2f}")

        # Write AUC to a JSON file
        with open(Path(self.output_dir) / 'auc.json', 'w') as file:
            json.dump({"auc_value": roc_auc}, file, indent=4)

    def calculate_tpr(self, tp, fn) -> float:
        """ Calculate True Positive Rate (Sensitivity or Recall) """
        try:
            value = tp / (tp + fn)
            if np.isnan(value):
                error_value = 0.0
                return error_value
            else:
                return value
        except ZeroDivisionError:
            return 0.0

    def calculate_fpr(self, fp, tn) -> float:
        """ Calculate False Positive Rate """
        try:
            return fp / (fp + tn)
        except ZeroDivisionError:
            return 1.0

    def calculate_tnr(self, tn, fp) -> float:
        """ Calculate True Negative Rate (Specificity) """
        try:
            return tn / (tn + fp)
        except ZeroDivisionError:
            return 0.0

    def calculate_fnr(self, fn, tp) -> float:
        """ Calculate False Negative Rate """
        try:
            return fn / (fn + tp)
        except ZeroDivisionError:
            return 1.0

    def calculate_acc(self, tp, tn, fp, fn) -> float:
        """ Calculate Accuracy """
        try:
            return (tp + tn) / (tp + tn + fp + fn)
        except ZeroDivisionError:
            return 0.0

    def calculate_bacc(self, tp, tn, fp, fn) -> float:
        """ Calculate Balanced Accuracy """
        sensitivity: float = self.calculate_tpr(tp, fn)
        specificity: float = self.calculate_tnr(tn, fp)
        return (sensitivity + specificity) / 2

    def calculate_mcc(self, tp, tn, fp, fn) -> float:
        """ Calculate Matthew's Correlation Coefficient """
        try:
            numerator = (tp * tn) - (fp * fn)
            denominator = np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))
            return numerator / denominator
        except ZeroDivisionError:
            return 0.0

    def calculate_f1(self, tp, tn, fp, fn):
        """ Calculate F1 Score """
        try:
            precision = tp / (tp + fp)
            recall = tp / (tp + fn)
            return 2 * (precision * recall) / (precision + recall)
        except ZeroDivisionError:
            return 0.0



import unittest

import numpy as np

from useckit.tests.test_utils import make_some_intelligent_noise
from useckit.util.dataset import Dataset
from useckit.util.dataset_windowsliced import WindowslicedDataset


class TestUseckit(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.x_train, self.y_train = make_some_intelligent_noise(labels=3, shape=(15, 10, 1))
        self.x_val, self.y_val = make_some_intelligent_noise(labels=3, shape=(15, 10, 1))
        self.x_test, self.y_test = make_some_intelligent_noise(labels=3, shape=(15, 10, 1))

    def test_dataset_initialization_train(self):
        exception_thrown = False
        try:
            data = Dataset(trainset_data=self.x_train,
                           trainset_labels=self.y_train)
        except ValueError:
            exception_thrown = True
        assert exception_thrown

    def test_dataset_initialization_trainval(self):
        data = Dataset(trainset_data=self.x_train,
                       trainset_labels=self.y_train,
                       validationset_data=self.x_val,
                       validationset_labels=self.y_val)
        return data

    def test_dataset_initialization_trainvalenroll(self):
        data = Dataset(trainset_data=self.x_train,
                       trainset_labels=self.y_train,
                       validationset_data=self.x_val,
                       validationset_labels=self.y_val,
                       testset_enrollment_data=self.x_test,
                       testset_enrollment_labels=self.y_test)
        return data

    def test_dataset_initialization_trainvalenrolltest(self):
        data = Dataset(trainset_data=self.x_train,
                       trainset_labels=self.y_train,
                       validationset_data=self.x_val,
                       validationset_labels=self.y_val,
                       testset_enrollment_data=self.x_test,
                       testset_enrollment_labels=self.y_test,
                       testset_matching_data=self.x_test,
                       testset_matching_labels=self.y_test)
        return data

    def test_dataset_initialization_traintest(self):
        data = Dataset(trainset_data=self.x_train,
                       trainset_labels=self.y_train,
                       testset_enrollment_data=self.x_test,
                       testset_enrollment_labels=self.y_test)
        return data

    def test_slicing(self):
        input_data, input_label = make_some_intelligent_noise(labels=3, shape=(15, 10, 1))
        data = WindowslicedDataset(window_slicing_size=4,
                                   window_slicing_stride=2,
                                   trainset_data=input_data,
                                   trainset_labels=input_label,
                                   testset_enrollment_data=input_data,
                                   testset_enrollment_labels=input_label,
                                   testset_matching_data=input_data,
                                   testset_matching_labels=input_label
                                   )
        assert data.trainset_data.shape == data.validationset_data.shape == data.testset_enrollment_data.shape == \
               data.testset_matching_data.shape == (60, 4, 1)
        assert data.trainset_labels.shape == data.validationset_labels.shape == \
               data.testset_enrollment_labels.shape == data.testset_matching_labels.shape == (60,)
        slicedsample_origin_target = np.array([[i, i, i, i] for i in range(15)], dtype=int).flatten()
        assert np.array_equal(data.trainset_slicedsample_origin, data.validationset_slicedsample_origin) and \
               np.array_equal(data.validationset_slicedsample_origin, data.testset_enrollment_slicedsample_origin) and \
               np.array_equal(data.testset_enrollment_slicedsample_origin,
                              data.testset_matching_slicedsample_origin) and \
               np.array_equal(data.testset_matching_slicedsample_origin, slicedsample_origin_target)

        for i in range(15):
            assert np.array_equal(data.trainset_sliceorigin_mask[i], (slicedsample_origin_target == i))


if __name__ == '__main__':
    unittest.main()



import numpy as np

from ._fixed_false_negative_thresholding_method import FixedFalseNegativeThresholdingMethod
from .anomaly_evaluation_method_base import AnomalyBaseEvaluationMethod, BaseThresholdingMethod
from ..prediction_models.anomaly_prediction_model_base import AnomalyBasePredictionModel
from ....evaluation.identification_with_reject import IdentificationOrRejectModel, \
    perform_identification_or_reject_evaluation
from ....util.dataset import Dataset


class AnomalyIdentificationOrRejectModel(IdentificationOrRejectModel):

    def __init__(self,
                 anomaly_prediction_model: AnomalyBasePredictionModel,
                 rejection_thresholds_per_trained_user: dict,
                 dataset: Dataset):
        self.anomaly_prediction_model = anomaly_prediction_model
        self.rejection_thresholds_per_trained_user = rejection_thresholds_per_trained_user
        self.dataset = dataset

    def identify_or_reject(self, samples: np.ndarray):
        preds, labels = self.anomaly_prediction_model.predict(samples)
        preds = preds.transpose()  # 0-axis = samples, 1-axis = the predictions of the different models for the sample

        result = np.argmin(preds, axis=1)  # reduce to the indices of the model with the minimum prediction
        result_index = np.argmin(preds, axis=1)  # and a copy, please
        for i in range(len(result)):
            result[i] = labels[result[i]]  # pick for each sample the label corresponding to the model with the
            # min prediction

        for i in range(len(preds)):  # iterate all samples
            reject = preds[i][result_index[i]] > self.rejection_thresholds_per_trained_user[labels[result_index[i]]]
            if reject:
                result[i] = -1
        return self.dataset.reverse_label_transform(result)


class IdentificationWithReject(AnomalyBaseEvaluationMethod):

    def __init__(self,
                 output_dir: str = "evaluation_identification_with_reject",
                 threshold_method: BaseThresholdingMethod = None):
        super().__init__(output_dir)
        self.threshold_method = threshold_method

    def evaluate(self, dataset: Dataset, prediction_model: AnomalyBasePredictionModel, **kwargs):
        if self.threshold_method is None:
            self.threshold_method = FixedFalseNegativeThresholdingMethod(prediction_model, self.output_dir)
        thresholds = self.threshold_method.compute_thresholds(dataset.testset_enrollment_data,
                                                              dataset.testset_enrollment_labels)
        identification_or_reject_model = AnomalyIdentificationOrRejectModel(prediction_model, thresholds, dataset)
        perform_identification_or_reject_evaluation(identification_or_reject_model, dataset, self.output_dir)



# ATTRIBUTION: this file originates from https://github.com/hfawaz/dl-4-tsc
# It was published on github.com by Ismail Fawaz, Hassan and Forestier, Germain and Weber, Jonathan and Idoumghar, Lhassane and Muller, Pierre-Alain
# for their work "Deep learning for time series classification: a review", in Data Mining and Knowledge Discovery.
# LICENSE is GNU GENERAL PUBLIC LICENSE version 3 (29 June 2007).
# Adopted by BLINDFORREVIEW for 'useckit'.
# Please cite:
"""
@article{IsmailFawaz2018deep,
  Title                    = {Deep learning for time series classification: a review},
  Author                   = {Ismail Fawaz, Hassan and Forestier, Germain and Weber, Jonathan and Idoumghar, Lhassane and Muller, Pierre-Alain},
  journal                  = {Data Mining and Knowledge Discovery},
  Year                     = {2019},
  volume                   = {33},
  number                   = {4},
  pages                    = {917--963},
}
"""
import os
import time

import numpy as np
# MCNN model
import tensorflow.keras as keras

from .tsc_prediction_model_base import TSCBasePredictionModel
from ....util.dataset import Dataset
from ....util.utils import calculate_metrics, create_directory


class dl4tsc_mcnn(TSCBasePredictionModel):

    def persist_model_trainables(self, path_like: str = 'saved_trainable_values', **kwargs):
        raise NotImplementedError

    def restore_model_trainables(self, path_like, dataset: Dataset = None, **kwargs):
        raise NotImplementedError

    def __init__(self, nb_classes, verbose=False, nb_epochs=None, output_dir: str = "MCNN_model_out"):
        super().__init__(verbose=verbose, output_dir=output_dir)
        self.nb_classes = nb_classes
        self.nb_epochs = nb_epochs
        self.pool_factors = [2, 3, 5]  # used for hyperparameters grid search
        self.filter_sizes = [0.05, 0.1, 0.2]  # used for hyperparameters grid search
        self.y_pred = None
        self.input_shapes = None
        self.ma_base, self.ma_step, self.ma_num = 5, 6, 1
        self.ds_base, self.ds_step, self.ds_num = 2, 1, 4

    def slice_data(self, data_x, data_y, slice_ratio):
        n = data_x.shape[0]
        length = data_x.shape[1]
        n_dim = data_x.shape[2]  # for MTS
        nb_classes = self.nb_classes

        length_sliced = int(length * slice_ratio)

        increase_num = length - length_sliced + 1  # if increase_num =5, it means one ori becomes 5 new instances.
        n_sliced = n * increase_num

        new_x = np.zeros((n_sliced, length_sliced, n_dim))
        new_y = np.zeros((n_sliced, nb_classes))
        for i in range(n):
            for j in range(increase_num):
                new_x[i * increase_num + j, :, :] = data_x[i, j: j + length_sliced, :]
                new_y[i * increase_num + j] = np.int_(data_y[i].astype(np.float32))

        return new_x, new_y

    def split_train(self, train_x, train_y):
        # shuffle for splitting train set and dataset
        n = train_x.shape[0]
        ind = np.arange(n)
        np.random.shuffle(ind)  # shuffle the train set

        # split train set into train set and validation set
        valid_x = train_x[ind[0:int(0.2 * n)]]
        valid_y = train_y[ind[0:int(0.2 * n)]]

        ind = np.delete(ind, (range(0, int(0.2 * n))))

        train_x = train_x[ind]
        train_y = train_y[ind]

        return train_x, train_y, valid_x, valid_y

    def _downsample(self, data_x, sample_rate, offset=0):
        num = data_x.shape[0]
        length_x = data_x.shape[1]
        num_dim = data_x.shape[2]  # for MTS
        last_one = 0
        if length_x % sample_rate > offset:
            last_one = 1
        new_length = int(np.floor(length_x / sample_rate)) + last_one
        output = np.zeros((num, new_length, num_dim))
        for i in range(new_length):
            output[:, i] = np.array(data_x[:, offset + sample_rate * i])

        return output

    def _movingavrg(self, data_x, window_size):
        num = data_x.shape[0]
        length_x = data_x.shape[1]
        num_dim = data_x.shape[2]  # for MTS
        output_len = length_x - window_size + 1
        output = np.zeros((num, output_len, num_dim))
        for i in range(output_len):
            output[:, i] = np.mean(data_x[:, i: i + window_size], axis=1)
        return output

    def movingavrg(self, data_x, window_base, step_size, num):
        if num == 0:
            return (None, [])
        out = self._movingavrg(data_x, window_base)
        data_lengths = [out.shape[1]]
        for i in range(1, num):
            window_size = window_base + step_size * i
            if window_size > data_x.shape[1]:
                continue
            new_series = self._movingavrg(data_x, window_size)
            data_lengths.append(new_series.shape[1])
            out = np.concatenate([out, new_series], axis=1)
        return (out, data_lengths)

    def batch_movingavrg(self, train, valid, test, window_base, step_size, num):
        (new_train, lengths) = self.movingavrg(train, window_base, step_size, num)
        (new_valid, lengths) = self.movingavrg(valid, window_base, step_size, num)
        (new_test, lengths) = self.movingavrg(test, window_base, step_size, num)
        return (new_train, new_valid, new_test, lengths)

    def downsample(self, data_x, base, step_size, num):
        # the case for dataset JapaneseVowels MTS
        if data_x.shape[1] == 26:
            return (None, [])  # too short to apply downsampling
        if num == 0:
            return (None, [])
        out = self._downsample(data_x, base, 0)
        data_lengths = [out.shape[1]]
        # for offset in range(1,base): #for the base case
        #    new_series = _downsample(data_x, base, offset)
        #    data_lengths.append( new_series.shape[1] )
        #    out = np.concatenate( [out, new_series], axis = 1)
        for i in range(1, num):
            sample_rate = base + step_size * i
            if sample_rate > data_x.shape[1]:
                continue
            for offset in range(0, 1):  # sample_rate):
                new_series = self._downsample(data_x, sample_rate, offset)
                data_lengths.append(new_series.shape[1])
                out = np.concatenate([out, new_series], axis=1)
        return (out, data_lengths)

    def batch_downsample(self, train, valid, test, window_base, step_size, num):
        (new_train, lengths) = self.downsample(train, window_base, step_size, num)
        (new_valid, lengths) = self.downsample(valid, window_base, step_size, num)
        (new_test, lengths) = self.downsample(test, window_base, step_size, num)
        return (new_train, new_valid, new_test, lengths)

    def get_pool_factor(self, conv_shape, pool_size):
        for pool_factor in self.pool_factors:
            temp_pool_size = int(int(conv_shape) / pool_factor)
            print(temp_pool_size)
            if temp_pool_size == pool_size:
                return pool_factor

        raise Exception('Error on pool factor')

    def train(self, x_train, y_train, x_val, y_val, x_test, y_test, y_true, pool_factor=None, filter_size=None,
              do_train=True):
        window_size = 0.2
        n_train_batch = 10
        n_epochs = 200 if self.nb_epochs is None else self.nb_epochs
        max_train_batch_size = 256

        ori_len = x_train.shape[1]  # original_length of time series
        slice_ratio = 0.9

        if do_train == True:
            kernel_size = int(ori_len * filter_size)

        if do_train == False:
            model = keras.models.load_model(os.path.join(self.output_directory, 'best_model.hdf5'))

            # model.summary()

            pool_size = model.get_layer('max_pooling1d_1').get_config()['pool_size'][0]

            conv_shape = model.get_layer('conv1d_1').output_shape[1]

            pool_factor = self.get_pool_factor(conv_shape, pool_size)

        # restrict slice ratio when data lenght is too large
        if ori_len > 500:
            slice_ratio = slice_ratio if slice_ratio > 0.98 else 0.98
        elif ori_len < 16:
            slice_ratio = 0.7

        increase_num = ori_len - int(ori_len * slice_ratio) + 1  # this can be used as the bath size

        train_batch_size = int(x_train.shape[0] * increase_num / n_train_batch)
        if train_batch_size > max_train_batch_size:
            # limit the train_batch_size 
            n_train_batch = int(x_train.shape[0] * increase_num / max_train_batch_size)

        # data augmentation by slicing the length of the series 
        x_train, y_train = self.slice_data(x_train, y_train, slice_ratio)
        x_val, y_val = self.slice_data(x_val, y_val, slice_ratio)
        x_test, y_test = self.slice_data(x_test, y_test, slice_ratio)

        train_set_x, train_set_y = x_train, y_train
        valid_set_x, valid_set_y = x_val, y_val
        test_set_x, _ = x_test, y_test

        valid_num = valid_set_x.shape[0]

        # print("increase factor is ", increase_num, ', ori len', ori_len)
        valid_num_batch = int(valid_num / increase_num)

        test_num = test_set_x.shape[0]
        test_num_batch = int(test_num / increase_num)

        length_train = train_set_x.shape[1]  # length after slicing.

        window_size = int(length_train * window_size) if window_size < 1 else int(window_size)

        # *******set up the ma and ds********#
        ma_base, ma_step, ma_num = self.ma_base, self.ma_step, self.ma_num
        ds_base, ds_step, ds_num = self.ds_base, self.ds_step, self.ds_num

        ds_num_max = length_train / (pool_factor * window_size)
        ds_num = int(min(ds_num, ds_num_max))
        self.ds_num = ds_num

        # *******set up the ma and ds********#

        (ma_train, ma_valid, ma_test, ma_lengths) = self.batch_movingavrg(train_set_x,
                                                                          valid_set_x, test_set_x,
                                                                          ma_base, ma_step, ma_num)
        (ds_train, ds_valid, ds_test, ds_lengths) = self.batch_downsample(train_set_x,
                                                                          valid_set_x, test_set_x,
                                                                          ds_base, ds_step, ds_num)

        # concatenate directly
        data_lengths = [length_train]
        # downsample part:
        if ds_lengths != []:
            data_lengths += ds_lengths
            train_set_x = np.concatenate([train_set_x, ds_train], axis=1)
            valid_set_x = np.concatenate([valid_set_x, ds_valid], axis=1)
            test_set_x = np.concatenate([test_set_x, ds_test], axis=1)

        # moving average part
        if ma_lengths != []:
            data_lengths += ma_lengths
            train_set_x = np.concatenate([train_set_x, ma_train], axis=1)
            valid_set_x = np.concatenate([valid_set_x, ma_valid], axis=1)
            test_set_x = np.concatenate([test_set_x, ma_test], axis=1)
        # print("Data length:", data_lengths)

        n_train_size = train_set_x.shape[0]
        n_valid_size = valid_set_x.shape[0]
        n_test_size = test_set_x.shape[0]
        batch_size = int(n_train_size / n_train_batch)
        n_train_batches = int(n_train_size / batch_size)
        data_dim = train_set_x.shape[1]
        num_dim = train_set_x.shape[2]  # For MTS
        nb_classes = self.nb_classes

        # print('train size', n_train_size, ',valid size', n_valid_size, ' test size', n_test_size)
        # print('batch size ', batch_size)
        # print('n_train_batches is ', n_train_batches)
        # print('data dim is ', data_dim)
        # print('---------------------------')

        ######################
        # BUILD ACTUAL MODEL #
        ######################
        # print('building the model...')

        input_shapes, max_length = self.get_list_of_input_shapes(data_lengths, num_dim)
        self.input_shapes = input_shapes

        start_time = time.time()

        best_validation_loss = np.inf

        if do_train == True:

            model = self.build_model(input_shapes, nb_classes, pool_factor, kernel_size)

            if (self.verbose == True):
                model.summary()

            # print('Training')

            # early-stopping parameters
            patience = 10000  # look as this many examples regardless
            patience_increase = 2  # wait this much longer when a new best is
            # found
            improvement_threshold = 0.995  # a relative improvement of this much is
            # considered significant
            validation_frequency = min(n_train_batches, patience / 2)
            # go through this many
            # minibatche before checking the network
            # on the validation set; in this case we
            # check every epoch
            max_before_stopping = 500

            best_iter = 0
            valid_loss = 0.

            epoch = 0
            done_looping = False
            num_no_update_epoch = 0
            epoch_avg_cost = float('inf')
            epoch_avg_err = float('inf')

            while (epoch < n_epochs) and (not done_looping):
                epoch = epoch + 1
                epoch_train_err = 0.
                epoch_cost = 0.

                num_no_update_epoch += 1
                if num_no_update_epoch == max_before_stopping:
                    break

                for minibatch_index in range(n_train_batches):

                    iteration = (epoch - 1) * n_train_batches + minibatch_index

                    x = train_set_x[minibatch_index * batch_size: (minibatch_index + 1) * batch_size]
                    y = train_set_y[minibatch_index * batch_size: (minibatch_index + 1) * batch_size]

                    x = self.split_input_for_model(x, input_shapes)

                    cost_ij, accuracy = model.train_on_batch(x, y)

                    train_err = 1 - accuracy

                    epoch_train_err = epoch_train_err + train_err
                    epoch_cost = epoch_cost + cost_ij

                    if (iteration + 1) % validation_frequency == 0:

                        valid_losses = []
                        for i in range(valid_num_batch):
                            x = valid_set_x[i * (increase_num): (i + 1) * (increase_num)]
                            y_pred = model.predict_on_batch(self.split_input_for_model(x, input_shapes))

                            # convert the predicted from binary to integer
                            y_pred = np.argmax(y_pred, axis=1)
                            label = np.argmax(valid_set_y[i * increase_num])

                            unique_value, sub_ind, correspond_ind, count = np.unique(y_pred, True, True, True)
                            unique_value = unique_value.tolist()

                            curr_err = 1.
                            if label in unique_value:
                                target_ind = unique_value.index(label)
                                count = count.tolist()
                                sorted_count = sorted(count)
                                if count[target_ind] == sorted_count[-1]:
                                    if len(sorted_count) > 1 and sorted_count[-1] == sorted_count[-2]:
                                        curr_err = 0.5  # tie
                                    else:
                                        curr_err = 0
                            valid_losses.append(curr_err)
                        valid_loss = sum(valid_losses) / float(len(valid_losses))

                        # print('...epoch%i,valid err: %.5f |' % (epoch,valid_loss))

                        # if we got the best validation score until now
                        if valid_loss <= best_validation_loss:
                            num_no_update_epoch = 0

                            # improve patience if loss improvement is good enough
                            if valid_loss < best_validation_loss * improvement_threshold:
                                patience = max(patience, iteration * patience_increase)

                            # save best validation score and iteration number
                            best_validation_loss = valid_loss
                            best_iter = iteration

                            # save model in h5 format
                            model.save(os.path.join(self.output_directory, 'best_model.hdf5'))

                        model.save(os.path.join(self.output_directory, 'last_model.hdf5'))
                    if patience <= iteration:
                        done_looping = True
                        break
                epoch_avg_cost = epoch_cost / n_train_batches
                epoch_avg_err = epoch_train_err / n_train_batches

                # print ('train err %.5f, cost %.4f' %(epoch_avg_err,epoch_avg_cost))
                if epoch_avg_cost == 0:
                    break

            # print('Optimization complete.')

        # test the model
        # print('Testing')
        # load best model
        model = keras.models.load_model(os.path.join(self.output_directory, 'best_model.hdf5'))

        # get the true predictions of the test set
        y_predicted = []
        for i in range(test_num_batch):
            x = test_set_x[i * (increase_num): (i + 1) * (increase_num)]
            y_pred = model.predict_on_batch(self.split_input_for_model(x, input_shapes))

            # convert the predicted from binary to integer 
            y_pred = np.argmax(y_pred, axis=1)

            unique_value, sub_ind, correspond_ind, count = np.unique(y_pred, True, True, True)

            idx_max = np.argmax(count)
            predicted_label = unique_value[idx_max]

            y_predicted.append(predicted_label)

        self.y_pred = y_pred = np.array(y_predicted)

        duration = time.time() - start_time

        df_metrics = calculate_metrics(y_true, y_pred, duration)

        # print(y_true.shape)
        # print(y_pred.shape)

        df_metrics.to_csv(os.path.join(self.output_directory, 'df_metrics.csv'), index=False)

        return df_metrics, model, best_validation_loss

    def split_input_for_model(self, x, input_shapes):
        res = []
        indx = 0
        for input_shape in input_shapes:
            res.append(x[:, indx:indx + input_shape[0], :])
            indx = indx + input_shape[0]
        return res

    def build_model(self, input_shapes, nb_classes, pool_factor, kernel_size):
        input_layers = []
        stage_1_layers = []

        for input_shape in input_shapes:
            input_layer = keras.layers.Input(input_shape)

            input_layers.append(input_layer)

            conv_layer = keras.layers.Conv1D(filters=256, kernel_size=kernel_size, padding='same',
                                             activation='sigmoid', kernel_initializer='glorot_uniform')(input_layer)

            # should all concatenated have the same length
            pool_size = int(int(conv_layer.shape[1]) / pool_factor)

            max_layer = keras.layers.MaxPooling1D(pool_size=pool_size)(conv_layer)

            # max_layer = keras.layers.GlobalMaxPooling1D()(conv_layer)

            stage_1_layers.append(max_layer)

        concat_layer = keras.layers.Concatenate(axis=-1)(stage_1_layers)

        kernel_size = int(min(kernel_size, int(concat_layer.shape[1])))  # kernel shouldn't exceed the length

        full_conv = keras.layers.Conv1D(filters=256, kernel_size=kernel_size, padding='same',
                                        activation='sigmoid', kernel_initializer='glorot_uniform')(concat_layer)

        pool_size = int(int(full_conv.shape[1]) / pool_factor)

        full_max = keras.layers.MaxPooling1D(pool_size=pool_size)(full_conv)

        full_max = keras.layers.Flatten()(full_max)

        fully_connected = keras.layers.Dense(units=256, activation='sigmoid',
                                             kernel_initializer='glorot_uniform')(full_max)

        output_layer = keras.layers.Dense(units=nb_classes, activation='softmax',
                                          kernel_initializer='glorot_uniform')(fully_connected)

        model = keras.models.Model(inputs=input_layers, outputs=output_layer)

        model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(learning_rate=0.1),
                      metrics=['accuracy'])

        return model

    def get_list_of_input_shapes(self, data_lengths, num_dim):
        input_shapes = []
        max_length = 0
        for i in data_lengths:
            input_shapes.append((i, num_dim))
            max_length = max(max_length, i)
        return input_shapes, max_length

    def fit(self, dataset: Dataset, *args, **kwargs):
        x_train, x_val, y_train, y_val, _, input_shape, nb_classes = self.convert_dataset_to_legacy_values(dataset)
        x_test = dataset.testset_matching_data
        _, _, _, y_test = dataset.view_one_hot_encoded_labels()
        y_true = dataset.testset_matching_labels

        best_df_metrics = None
        best_valid_loss = np.inf

        output_directory_root = self.output_dir
        # grid search 
        for pool_factor in self.pool_factors:
            for filter_size in self.filter_sizes:
                self.output_directory = os.path.join(output_directory_root,
                                                     'hyper_param_search',
                                                     'pool_factor_' + str(pool_factor),
                                                     'filter_size_' + str(filter_size))
                create_directory(self.output_directory)
                df_metrics, model, valid_loss = self.train(x_train, y_train, x_val,
                                                           y_val, x_test, y_test, y_true, pool_factor, filter_size)

                if (valid_loss < best_valid_loss):
                    best_valid_loss = valid_loss
                    best_df_metrics = df_metrics
                    best_df_metrics.to_csv(os.path.join(output_directory_root, 'df_metrics.csv'), index=False)
                    model.save(os.path.join(output_directory_root, 'best_model.hdf5'))

                model = None
                # clear memeory 
                keras.backend.clear_session()

    def predict(self, x_test, y_true, x_train, y_train, y_test, return_df_metrics=False):
        if return_df_metrics:
            raise NotImplementedError('returning df_metrics is not supported for MCNN prediction models!')
        else:
            model = keras.models.load_model(os.path.join(self.output_directory, 'best_model.hdf5'))
            (ma_test, ma_lengths) = self.movingavrg(x_test, self.ma_base, self.ma_step, self.ma_num)
            (ds_test, ds_lengths) = self.downsample(x_test, self.ds_base, self.ds_step, self.ds_num)
            if ds_lengths != []:
                x_test = np.concatenate([x_test, ds_test], axis=1)

            # moving average part
            if ma_lengths != []:
                x_test = np.concatenate([x_test, ma_test], axis=1)

            y_pred = model.predict_on_batch(self.split_input_for_model(x_test, self.input_shapes))
            y_pred = np.argmax(y_pred, axis=1)
            return y_pred



import setuptools


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="useckit",
    version="0.5.0a10",
    author="BlindForReview.com",
    author_email="mail@blindforreview.com",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://blindforreview.com",
    packages=setuptools.find_packages(include=['useckit', 'useckit.*']),
    data_files=[('examples', ['examples/useckit-high-level-v2.ipynb'])],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=parse_requirements('requirements.txt')
)



# ATTRIBUTION: this file originates from https://github.com/hfawaz/dl-4-tsc
# It was published on github.com by Ismail Fawaz, Hassan and Forestier, Germain and Weber, Jonathan and Idoumghar, Lhassane and Muller, Pierre-Alain
# for their work "Deep learning for time series classification: a review", in Data Mining and Knowledge Discovery.
# LICENSE is GNU GENERAL PUBLIC LICENSE version 3 (29 June 2007).
# Adopted by BLINDFORREVIEW for 'useckit'.
# Please cite:
"""
@article{IsmailFawaz2018deep,
  Title                    = {Deep learning for time series classification: a review},
  Author                   = {Ismail Fawaz, Hassan and Forestier, Germain and Weber, Jonathan and Idoumghar, Lhassane and Muller, Pierre-Alain},
  journal                  = {Data Mining and Knowledge Discovery},
  Year                     = {2019},
  volume                   = {33},
  number                   = {4},
  pages                    = {917--963},
}
"""
import os

import tensorflow.keras as keras

from .classification_keras_prediction_model import ClassificationKerasPredictionModel


class dl4tsc_inception(ClassificationKerasPredictionModel):

    def __init__(self, verbose=False,
                 batch_size=lambda x_train_shape_0: int(min(x_train_shape_0 / 10, 16)),
                 lr=0.001,
                 nb_filters=32,
                 use_residual=True,
                 use_bottleneck=True,
                 depth=6,
                 kernel_size=41,
                 nb_epochs=1500,
                 class_weights=None,
                 output_dir: str = "inception_model_out"):
        super().__init__(output_dir=output_dir, verbose=verbose, nb_epochs=nb_epochs, class_weights=class_weights,
                         batch_size=batch_size)

        self.nb_filters = nb_filters
        self.use_residual = use_residual
        self.use_bottleneck = use_bottleneck
        self.depth = depth
        self.kernel_size = kernel_size - 1
        self.callbacks = None

        self.bottleneck_size = 32
        self.lr = lr

    def _inception_module(self, input_tensor, stride=1, activation='linear'):

        if self.use_bottleneck and int(input_tensor.shape[-1]) > self.bottleneck_size:
            input_inception = keras.layers.Conv1D(filters=self.bottleneck_size, kernel_size=1,
                                                  padding='same', activation=activation, use_bias=False)(input_tensor)
        else:
            input_inception = input_tensor

        # kernel_size_s = [3, 5, 8, 11, 17]
        kernel_size_s = [self.kernel_size // (2 ** i) for i in range(3)]

        conv_list = []

        for i in range(len(kernel_size_s)):
            conv_list.append(keras.layers.Conv1D(filters=self.nb_filters, kernel_size=kernel_size_s[i],
                                                 strides=stride, padding='same', activation=activation, use_bias=False)(
                input_inception))

        max_pool_1 = keras.layers.MaxPool1D(pool_size=3, strides=stride, padding='same')(input_tensor)

        conv_6 = keras.layers.Conv1D(filters=self.nb_filters, kernel_size=1,
                                     padding='same', activation=activation, use_bias=False)(max_pool_1)

        conv_list.append(conv_6)

        x = keras.layers.Concatenate(axis=2)(conv_list)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Activation(activation='relu')(x)
        return x

    def _shortcut_layer(self, input_tensor, out_tensor):
        shortcut_y = keras.layers.Conv1D(filters=int(out_tensor.shape[-1]), kernel_size=1,
                                         padding='same', use_bias=False)(input_tensor)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

        x = keras.layers.Add()([shortcut_y, out_tensor])
        x = keras.layers.Activation('relu')(x)
        return x

    def build_model(self, input_shape, nb_classes):
        input_layer = keras.layers.Input(input_shape)

        x = input_layer
        input_res = input_layer

        for d in range(self.depth):

            x = self._inception_module(x)

            if self.use_residual and d % 3 == 2:
                x = self._shortcut_layer(input_res, x)
                input_res = x

        gap_layer = keras.layers.GlobalAveragePooling1D()(x)

        output_layer = keras.layers.Dense(nb_classes, activation='softmax')(gap_layer)

        model = keras.models.Model(inputs=input_layer, outputs=output_layer)

        model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(self.lr),
                      metrics=['accuracy'])

        reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50,
                                                      min_lr=0.0001)

        file_path = os.path.join(self.output_dir, 'best_model.weights.h5')

        model_checkpoint = keras.callbacks.ModelCheckpoint(filepath=file_path, monitor='val_loss',
                                                           save_best_only=True)

        self.callbacks = [reduce_lr, model_checkpoint]

        return model



from abc import abstractmethod, ABC

import numpy as np

from ..prediction_models.distance_prediction_model_base import DistanceBasePredictionModel
from ..._paradigm_base import EvaluationMethodBase
from ....util.dataset import Dataset


class BaseThresholdingMethod(ABC):

    @abstractmethod
    def compute_threshold(self, enrollment_data: np.ndarray, enrollment_labels: np.ndarray) -> float:
        pass


class DistanceBaseEvaluationMethod(EvaluationMethodBase):

    @abstractmethod
    def evaluate(self, dataset: Dataset, prediction_model: DistanceBasePredictionModel, **kwargs):
        pass
